import unittest
import numpy as np
from chalmers_qubit.utils.randomized_benchmarking.clifford_group import SingleQubitClifford, TwoQubitClifford

class TestClifford(unittest.TestCase):

    def setUp(self):
        # Clear caches before tests
        SingleQubitClifford.CLIFFORD_HASH_TABLE.clear()
        TwoQubitClifford.CLIFFORD_HASH_TABLE.clear()
        TwoQubitClifford._PTM_CACHE.clear()
        TwoQubitClifford._GATE_DECOMP_CACHE.clear()
        
        # Define test cases as CliffordClass
        self.test_cases = [SingleQubitClifford, TwoQubitClifford]


    def test_clifford_equality(self):
        for CliffordClass in self.test_cases:
            with self.subTest(CliffordClass=CliffordClass.__name__):
                c1 = CliffordClass(5)
                c2 = CliffordClass(5)
                c3 = CliffordClass(6)

                self.assertEqual(c1, c2, f"{CliffordClass.__name__}s with the same index should be equal")
                self.assertNotEqual(c1, c3, f"{CliffordClass.__name__}s with different indices should not be equal")

    def test_identity_clifford(self):
        for CliffordClass in self.test_cases:
            with self.subTest(CliffordClass=CliffordClass.__name__):
                ptm = CliffordClass(0).pauli_transfer_matrix
                size = ptm.shape[0]
                eye = np.identity(size)
                self.assertTrue(np.array_equal(eye, ptm), f"The identity {CliffordClass.__name__} should have an identity PTM")

    def test_get_inverse(self):
        for CliffordClass in self.test_cases:
            with self.subTest(CliffordClass=CliffordClass.__name__):
                idx = 10 # arbitrary index
                c = CliffordClass(idx)
                c_inv = c.get_inverse()
                # Check that the inverse is correct up to a global-phase
                c_eye = c_inv * c # this should be the identity
                eye = CliffordClass(0) # This should be the identity
                self.assertTrue(c_eye == eye, f"The product of a {CliffordClass.__name__} and its inverse should be the identity")

    def test_two_qubit_clifford_caching(self):
        """Test that the PTM and gate decomposition caches are shared between instances. 
        This is only used for TwoQubitClifford, as SingleQubitClifford does not have a cache.
        """
        CliffordClass = TwoQubitClifford

        idx1, idx2 = 3, 5 # arbitrary indices
        
        # Compute PTMs
        c1_ptm = CliffordClass(idx=idx1).pauli_transfer_matrix
        self.assertIn(idx1, CliffordClass._PTM_CACHE)
        np.testing.assert_array_equal(c1_ptm, CliffordClass._PTM_CACHE[idx1])
        
        # Check that the instances share the same cache
        c2 = CliffordClass(idx=idx2)
        np.testing.assert_array_equal(c2._PTM_CACHE[idx1], c1_ptm)
        
        # Compute gate decomposition
        decomp1 = CliffordClass(idx=idx1).gate_decomposition
        self.assertIn(idx1, CliffordClass._GATE_DECOMP_CACHE)
        self.assertEqual(decomp1, CliffordClass._GATE_DECOMP_CACHE[idx1])
        
        # Check that a new instance shares the same cache
        decomp2 = CliffordClass(idx=idx2)
        self.assertEqual(decomp2._GATE_DECOMP_CACHE[idx1], decomp1)

    def test_hash_table_generation(self):
        for CliffordClass in self.test_cases:
            with self.subTest(CliffordClass=CliffordClass.__name__):
                idx = 5 # Some arbitrary Clifford idx
                c = CliffordClass(idx)
                ptm = c.pauli_transfer_matrix

                # Trigger hash table population
                c.find_clifford_index(ptm)

                # Check if the hash table was correctly populated
                hash_value = c._hash_matrix(ptm)
                # Assert that the hash value is in the hash table
                self.assertIn(hash_value, CliffordClass.CLIFFORD_HASH_TABLE)
                # Assert that the index in the hash table matches the original index
                self.assertEqual(CliffordClass.CLIFFORD_HASH_TABLE[hash_value], idx)


if __name__ == '__main__':
    unittest.main()


# Modifications copyright (c) 2025 Pontus Vikstål
#
# MIT License
# 
# Copyright (c) 2016 DiCarlo lab-QuTech-Delft University of Technology
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from chalmers_qubit.utils.randomized_benchmarking.pauli_transfer_matrices import I, X, Y, Z, S, S2, H, CZ
from typing import List, Tuple, Dict, ClassVar

# The decomposition of the single qubit clifford group as per
# Epstein et al. Phys. Rev. A 89, 062321 (2014)
# Note: Explicitly reversing order because order of operators is order in time
epstein_efficient_decomposition = [[] for _ in range(24)]
epstein_efficient_decomposition[0] = ["I"]
epstein_efficient_decomposition[1] = ["Y90", "X90"]
epstein_efficient_decomposition[2] = ["mX90", "mY90"]
epstein_efficient_decomposition[3] = ["X180"]
epstein_efficient_decomposition[4] = ["mY90", "mX90"]
epstein_efficient_decomposition[5] = ["X90", "mY90"]
epstein_efficient_decomposition[6] = ["Y180"]
epstein_efficient_decomposition[7] = ["mY90", "X90"]
epstein_efficient_decomposition[8] = ["X90", "Y90"]
epstein_efficient_decomposition[9] = ["X180", "Y180"]
epstein_efficient_decomposition[10] = ["Y90", "mX90"]
epstein_efficient_decomposition[11] = ["mX90", "Y90"]
epstein_efficient_decomposition[12] = ["Y90", "X180"]
epstein_efficient_decomposition[13] = ["mX90"]
epstein_efficient_decomposition[14] = ["X90", "mY90", "mX90"]
epstein_efficient_decomposition[15] = ["mY90"]
epstein_efficient_decomposition[16] = ["X90"]
epstein_efficient_decomposition[17] = ["X90", "Y90", "X90"]
epstein_efficient_decomposition[18] = ["mY90", "X180"]
epstein_efficient_decomposition[19] = ["X90", "Y180"]
epstein_efficient_decomposition[20] = ["X90", "mY90", "X90"]
epstein_efficient_decomposition[21] = ["Y90"]
epstein_efficient_decomposition[22] = ["mX90", "Y180"]
epstein_efficient_decomposition[23] = ["X90", "Y90", "mX90"]

# The single qubit clifford group where each element is a 4x4 pauli transfer matrix
# Note: Explictly reversing order because order of operators is order in time
C1 = [np.empty([4, 4])] * (24)
C1[0] = np.linalg.multi_dot([I, I, I][::-1])
C1[1] = np.linalg.multi_dot([I, I, S][::-1])
C1[2] = np.linalg.multi_dot([I, I, S2][::-1])
C1[3] = np.linalg.multi_dot([X, I, I][::-1])
C1[4] = np.linalg.multi_dot([X, I, S][::-1])
C1[5] = np.linalg.multi_dot([X, I, S2][::-1])
C1[6] = np.linalg.multi_dot([Y, I, I][::-1])
C1[7] = np.linalg.multi_dot([Y, I, S][::-1])
C1[8] = np.linalg.multi_dot([Y, I, S2][::-1])
C1[9] = np.linalg.multi_dot([Z, I, I][::-1])
C1[10] = np.linalg.multi_dot([Z, I, S][::-1])
C1[11] = np.linalg.multi_dot([Z, I, S2][::-1])
C1[12] = np.linalg.multi_dot([I, H, I][::-1])
C1[13] = np.linalg.multi_dot([I, H, S][::-1])
C1[14] = np.linalg.multi_dot([I, H, S2][::-1])
C1[15] = np.linalg.multi_dot([X, H, I][::-1])
C1[16] = np.linalg.multi_dot([X, H, S][::-1])
C1[17] = np.linalg.multi_dot([X, H, S2][::-1])
C1[18] = np.linalg.multi_dot([Y, H, I][::-1])
C1[19] = np.linalg.multi_dot([Y, H, S][::-1])
C1[20] = np.linalg.multi_dot([Y, H, S2][::-1])
C1[21] = np.linalg.multi_dot([Z, H, I][::-1])
C1[22] = np.linalg.multi_dot([Z, H, S][::-1])
C1[23] = np.linalg.multi_dot([Z, H, S2][::-1])

# used in the CNOT-, iSWAP-, and SWAP-like gates
X90 = C1[16]
Y90 = C1[21]
mY90 = C1[15]

# The S1 exchange group is a subgroup of C1 (single qubit Clifford group) 
# and is used when generating C2 (two qubit Clifford group)
S1 = [
    C1[0],
    C1[1],
    C1[2],
]

class Clifford:
    """
    Base class for Clifford operations.

    Abstract base class for all Clifford operations.

    Attributes
    ----------
    idx : int
        Index of the Clifford operation.
    GROUP_SIZE : ClassVar[int]
        Size of the Clifford group.
    CLIFFORD_HASH_TABLE : ClassVar[Dict[int, int]]
        Hash table for fast lookup of Clifford indices.
    """

    CLIFFORD_HASH_TABLE: ClassVar[Dict[int, int]]
    GROUP_SIZE: ClassVar[int]

    def __init__(self, idx: int) -> None:
        """
        Initialize the Clifford object with a given index.

        Parameters
        ----------
        idx : int
            Index of the Clifford operation.

        Raises
        ------
        ValueError
            If the index is not valid (0 <= idx < GROUP_SIZE).
        """
        if not 0 <= idx < self.GROUP_SIZE:
            raise ValueError(f"Invalid Clifford index: {idx}. Must be 0 <= idx < {self.GROUP_SIZE}")
        self.idx = idx

    def __mul__(self, other: "Clifford") -> "Clifford":
        """
        Multiply two Clifford operations.

        Parameters
        ----------
        other : Clifford
            Another Clifford operation.

        Returns
        -------
        Clifford
            The product of the two Clifford Pauli Transfer Matrices.

        Raises
        ------
        TypeError
            If other is not the same type of Clifford.
        """
        if not isinstance(other, self.__class__):
            raise TypeError(f"Cannot multiply {self.__class__.__name__} with {other.__class__.__name__}")

        net_op = np.dot(self.pauli_transfer_matrix, other.pauli_transfer_matrix)
        idx = self.find_clifford_index(net_op)
        return self.__class__(idx)

    def __eq__(self, other: "Clifford") -> bool:
        """
        Check if two Clifford operations are equal.

        Parameters
        ----------
        other : Clifford
            Another Clifford operation.

        Returns
        -------
        bool
            True if the operations are equal, False otherwise.
        """
        if not isinstance(other, self.__class__):
            return False
        return self.idx == other.idx

    def get_inverse(self) -> "Clifford":
        """
        Get the inverse of this Clifford operation.

        Returns
        -------
        Clifford
            The inverse operation.
        """
        inverse_ptm = np.linalg.inv(self.pauli_transfer_matrix).astype(int)
        idx = self.find_clifford_index(inverse_ptm)
        return self.__class__(idx)

    @property
    def pauli_transfer_matrix(self) -> np.ndarray:
        """
        Returns the Pauli transfer matrix of the Clifford operation.

        Returns
        -------
        np.ndarray
            The Pauli transfer matrix.

        Raises
        ------
        NotImplementedError
            If not implemented in subclass.
        """
        raise NotImplementedError("Subclasses must implement pauli_transfer_matrix")

    @property
    def gate_decomposition(self) -> List[Tuple[List[str], str]]:
        """
        Returns the gate decomposition of the Clifford gate.

        Returns
        -------
        list of tuple
            Gate decomposition as a list of tuples (gate_name, qubit_identifier).

        Raises
        ------
        NotImplementedError
            If not implemented in subclass.
        """
        raise NotImplementedError("Subclasses must implement gate_decomposition")

    @classmethod
    def find_clifford_index(cls, matrix: np.ndarray) -> int:
        """
        Find the index of a Clifford matrix using hash lookup.

        Parameters
        ----------
        matrix : np.ndarray
            The Pauli transfer matrix.

        Returns
        -------
        int
            The index of the Clifford operation.

        Raises
        ------
        ValueError
            If the Clifford index is not found.
        """
        if not cls.CLIFFORD_HASH_TABLE:
            for idx in range(cls.GROUP_SIZE):
                ptm = cls(idx=idx).pauli_transfer_matrix
                hash_value = cls._hash_matrix(ptm)
                cls.CLIFFORD_HASH_TABLE[hash_value] = idx

        hash_value = cls._hash_matrix(matrix)
        if hash_value in cls.CLIFFORD_HASH_TABLE:
            return cls.CLIFFORD_HASH_TABLE[hash_value]

        raise ValueError("Clifford index not found.")

    @staticmethod
    def _hash_matrix(matrix: np.ndarray) -> int:
        """
        Create a hash value for a matrix using NumPy's internal representation.

        Uses the byte representation of the rounded integer matrix.

        Parameters
        ----------
        matrix : np.ndarray
            The Pauli transfer matrix to hash.

        Returns
        -------
        int
            Hash value of the Pauli transfer matrix.
        """
        return hash(matrix.round().astype(int).tobytes())

    def __repr__(self) -> str:
        """
        Return a string representation of the Clifford object.

        Returns
        -------
        str
            String representation.
        """
        return f"{self.__class__.__name__}(idx={self.idx})"


class SingleQubitClifford(Clifford):
    """
    Single Qubit Clifford gate class.

    The decomposition of the single qubit clifford group follows
    Epstein et al. Phys. Rev. A 89, 062321 (2014).

    Attributes
    ----------
    CLIFFORD_HASH_TABLE : dict
        Hash table for fast lookup.
    GROUP_SIZE : int
        Size of the single qubit Clifford group.
    """

    CLIFFORD_HASH_TABLE = {}
    GROUP_SIZE = 24

    @property
    def pauli_transfer_matrix(self) -> np.ndarray:
        """
        Returns the Pauli transfer matrix of the single qubit Clifford operation.

        Returns
        -------
        np.ndarray
            Pauli transfer matrix.
        """
        return C1[self.idx]

    @property
    def gate_decomposition(self) -> List[Tuple[List[str], str]]:
        """
        Returns the gate decomposition of the single qubit Clifford group
        according to the decomposition by Epstein et al.

        Returns
        -------
        list of tuple
            Each tuple contains (gate_name, qubit_identifier).
        """
        gate_decomp = [(g, "q0") for g in epstein_efficient_decomposition[self.idx]]
        return gate_decomp


class TwoQubitClifford(Clifford):
    """
    Two Qubit Clifford gate class.

    The Clifford decomposition closely follows:
    1. Corcoles et al. Process verification ... Phys. Rev. A. 2013.
    2. Barends et al. Superconducting quantum circuits ... Nature 2014.

    The two qubit Clifford group (C2) consists of 11520 two-qubit Cliffords,
    subdivided into four classes.

    Attributes
    ----------
    CLIFFORD_HASH_TABLE : dict
        Hash table for fast lookup.
    _PTM_CACHE : dict
        Cache for PTMs.
    _GATE_DECOMP_CACHE : dict
        Cache for gate decompositions.
    GROUP_SIZE : int
        Size of the two-qubit Clifford group.
    """

    CLIFFORD_HASH_TABLE = {}
    _PTM_CACHE = {}
    _GATE_DECOMP_CACHE = {}

    GROUP_SIZE_CLIFFORD = 24
    GROUP_SIZE_SINGLE_QUBIT = GROUP_SIZE_CLIFFORD**2
    GROUP_SIZE_S1 = 3
    GROUP_SIZE_CNOT = GROUP_SIZE_SINGLE_QUBIT * GROUP_SIZE_S1**2
    GROUP_SIZE_ISWAP = GROUP_SIZE_CNOT
    GROUP_SIZE_SWAP = GROUP_SIZE_SINGLE_QUBIT
    GROUP_SIZE = GROUP_SIZE_SINGLE_QUBIT + GROUP_SIZE_CNOT + GROUP_SIZE_ISWAP + GROUP_SIZE_SWAP

    assert GROUP_SIZE_SINGLE_QUBIT == 576
    assert GROUP_SIZE_CNOT == 5184
    assert GROUP_SIZE == 11_520

    @property
    def pauli_transfer_matrix(self) -> np.ndarray:
        """
        Returns the Pauli transfer matrix for the two-qubit Clifford operation.

        Returns
        -------
        np.ndarray
            Pauli transfer matrix.
        """
        if self.idx not in self._PTM_CACHE:
            if self.idx < 576:
                ptm = self.single_qubit_like_PTM(self.idx)
            elif self.idx < 576 + 5184:
                ptm = self.CNOT_like_PTM(self.idx - 576)
            elif self.idx < 576 + 2 * 5184:
                ptm = self.iSWAP_like_PTM(self.idx - (576 + 5184))
            else:
                ptm = self.SWAP_like_PTM(self.idx - (576 + 2 * 5184))
            self._PTM_CACHE[self.idx] = ptm
        return self._PTM_CACHE[self.idx]

    @property
    def gate_decomposition(self):
        """
        Returns the gate decomposition of the two qubit Clifford group.

        Returns
        -------
        list of tuple
            Gate decomposition as a list of tuples (gate_name, qubit_identifier).
        """
        if self.idx not in self._GATE_DECOMP_CACHE:
            if self.idx < 576:
                gate_decomp = self.single_qubit_like_gates(self.idx)
            elif self.idx < 576 + 5184:
                gate_decomp = self.CNOT_like_gates(self.idx - 576)
            elif self.idx < 576 + 2 * 5184:
                gate_decomp = self.iSWAP_like_gates(self.idx - (576 + 5184))
            else:
                gate_decomp = self.SWAP_like_gates(self.idx - (576 + 2 * 5184))
            self._GATE_DECOMP_CACHE[self.idx] = gate_decomp
        return self._GATE_DECOMP_CACHE[self.idx]

    @classmethod
    def single_qubit_like_PTM(cls, idx: int) -> np.ndarray:
        """
        Returns the pauli transfer matrix for gates of the single qubit like class.

        Parameters
        ----------
        idx : int
            Index within the single qubit like class.

        Returns
        -------
        np.ndarray
            Pauli transfer matrix.
        """
        assert idx < cls.GROUP_SIZE_SINGLE_QUBIT
        idx_q0 = idx % 24
        idx_q1 = idx // 24
        pauli_transfer_matrix = np.kron(C1[idx_q1], C1[idx_q0])
        return pauli_transfer_matrix

    @classmethod
    def single_qubit_like_gates(cls, idx: int) -> List[Tuple[str, str]]:
        """
        Returns the gates for Cliffords of the single qubit like class.

        Parameters
        ----------
        idx : int
            Index within the single qubit like class.

        Returns
        -------
        list of tuple
            Each tuple contains (gate_name, qubit_identifier).
        """
        assert idx < cls.GROUP_SIZE_SINGLE_QUBIT
        idx_q0 = idx % 24
        idx_q1 = idx // 24

        g_q0 = [(g, "q0") for g in epstein_efficient_decomposition[idx_q0]]
        g_q1 = [(g, "q1") for g in epstein_efficient_decomposition[idx_q1]]
        gates = g_q0 + g_q1
        return gates

    @classmethod
    def CNOT_like_PTM(cls, idx: int) -> np.ndarray:
        """
        Returns the pauli transfer matrix for gates of the CNOT-like class.

        Parameters
        ----------
        idx : int
            Index within the CNOT-like class.

        Returns
        -------
        np.ndarray
            Pauli transfer matrix.
        """
        assert idx < cls.GROUP_SIZE_CNOT
        idx_0 = idx % 24
        idx_1 = (idx // 24) % 24
        idx_2 = (idx // 576) % 3
        idx_3 = idx // 1728

        C1_q0 = np.kron(np.eye(4), C1[idx_0])
        C1_q1 = np.kron(C1[idx_1], np.eye(4))
        S1_q0 = np.kron(np.eye(4), S1[idx_2])
        S1y_q1 = np.kron(np.dot(C1[idx_3], Y90), np.eye(4))
        return np.linalg.multi_dot(list(reversed([C1_q0, C1_q1, CZ, S1_q0, S1y_q1])))

    @classmethod
    def CNOT_like_gates(cls, idx: int):
        """
        Returns the gates for Cliffords of the CNOT-like class.

        Parameters
        ----------
        idx : int
            Index within the CNOT-like class.

        Returns
        -------
        list of tuple
            Each tuple contains (gate_name, qubit_identifier).
        """
        assert idx < cls.GROUP_SIZE_CNOT
        idx_0 = idx % 24
        idx_1 = (idx // 24) % 24
        idx_2 = (idx // 576) % 3
        idx_3 = idx // 1728

        C1_q0 = [(g, "q0") for g in epstein_efficient_decomposition[idx_0]]
        C1_q1 = [(g, "q1") for g in epstein_efficient_decomposition[idx_1]]
        CZ = [("CZ", ["q0", "q1"])]

        idx_2s = SingleQubitClifford.find_clifford_index(S1[idx_2])
        S1_q0 = [(g, "q0") for g in epstein_efficient_decomposition[idx_2s]]
        idx_3s = SingleQubitClifford.find_clifford_index(np.dot(C1[idx_3], Y90))
        S1_yq1 = [(g, "q1") for g in epstein_efficient_decomposition[idx_3s]]

        gates = C1_q0 + C1_q1 + CZ + S1_q0 + S1_yq1
        return gates

    @classmethod
    def iSWAP_like_PTM(cls, idx: int) -> np.ndarray:
        """
        Returns the pauli transfer matrix for gates of the iSWAP-like class.

        Parameters
        ----------
        idx : int
            Index within the iSWAP-like class.

        Returns
        -------
        np.ndarray
            Pauli transfer matrix.
        """
        assert idx < cls.GROUP_SIZE_ISWAP
        idx_0 = idx % 24
        idx_1 = (idx // 24) % 24
        idx_2 = (idx // 576) % 3
        idx_3 = idx // 1728

        C1_q0 = np.kron(np.eye(4), C1[idx_0])
        C1_q1 = np.kron(C1[idx_1], np.eye(4))
        sq_swap_gates = np.kron(mY90, Y90)
        S1_q0 = np.kron(np.eye(4), np.dot(S1[idx_2], Y90))
        S1y_q1 = np.kron(np.dot(C1[idx_3], X90), np.eye(4))

        return np.linalg.multi_dot(
            list(reversed([C1_q0, C1_q1, CZ, sq_swap_gates, CZ, S1_q0, S1y_q1]))
        )

    @classmethod
    def iSWAP_like_gates(cls, idx: int):
        """
        Returns the gates for Cliffords of the iSWAP-like class.

        Parameters
        ----------
        idx : int
            Index within the iSWAP-like class.

        Returns
        -------
        list of tuple
            Each tuple contains (gate_name, qubit_identifier).
        """
        assert idx < cls.GROUP_SIZE_ISWAP
        idx_0 = idx % 24
        idx_1 = (idx // 24) % 24
        idx_2 = (idx // 576) % 3
        idx_3 = idx // 1728

        C1_q0 = [(g, "q0") for g in epstein_efficient_decomposition[idx_0]]
        C1_q1 = [(g, "q1") for g in epstein_efficient_decomposition[idx_1]]
        CZ = [("CZ", ["q0", "q1"])]

        sqs_idx_q0 = SingleQubitClifford.find_clifford_index(Y90)
        sqs_idx_q1 = SingleQubitClifford.find_clifford_index(mY90)
        sq_swap_gates_q0 = [(g, "q0") for g in epstein_efficient_decomposition[sqs_idx_q0]]
        sq_swap_gates_q1 = [(g, "q1") for g in epstein_efficient_decomposition[sqs_idx_q1]]

        idx_2s = SingleQubitClifford.find_clifford_index(np.dot(S1[idx_2], Y90))
        S1_q0 = [(g, "q0") for g in epstein_efficient_decomposition[idx_2s]]
        idx_3s = SingleQubitClifford.find_clifford_index(np.dot(C1[idx_3], X90))
        S1y_q1 = [(g, "q1") for g in epstein_efficient_decomposition[idx_3s]]

        gates = (
            C1_q0
            + C1_q1
            + CZ
            + sq_swap_gates_q0
            + sq_swap_gates_q1
            + CZ
            + S1_q0
            + S1y_q1
        )
        return gates

    @classmethod
    def SWAP_like_PTM(cls, idx: int) -> np.ndarray:
        """
        Returns the pauli transfer matrix for gates of the SWAP-like class.

        Parameters
        ----------
        idx : int
            Index within the SWAP-like class.

        Returns
        -------
        np.ndarray
            Pauli transfer matrix.
        """
        assert idx < cls.GROUP_SIZE_SWAP
        idx_q0 = idx % 24
        idx_q1 = idx // 24
        sq_like_cliff = np.kron(C1[idx_q1], C1[idx_q0])
        sq_swap_gates_0 = np.kron(Y90, mY90)
        sq_swap_gates_1 = np.kron(mY90, Y90)
        sq_swap_gates_2 = np.kron(Y90, np.eye(4))

        return np.linalg.multi_dot(
            list(
                reversed(
                    [
                        sq_like_cliff,
                        CZ,
                        sq_swap_gates_0,
                        CZ,
                        sq_swap_gates_1,
                        CZ,
                        sq_swap_gates_2,
                    ]
                )
            )
        )

    @classmethod
    def SWAP_like_gates(cls, idx: int):
        """
        Returns the gates for Cliffords of the SWAP-like class.

        Parameters
        ----------
        idx : int
            Index within the SWAP-like class.

        Returns
        -------
        list of tuple
            Each tuple contains (gate_name, qubit_identifier).
        """
        assert idx < cls.GROUP_SIZE_SWAP
        idx_q0 = idx % 24
        idx_q1 = idx // 24
        C1_q0 = [(g, "q0") for g in epstein_efficient_decomposition[idx_q0]]
        C1_q1 = [(g, "q1") for g in epstein_efficient_decomposition[idx_q1]]
        CZ = [("CZ", ["q0", "q1"])]

        sqs_idx_q0 = SingleQubitClifford.find_clifford_index(mY90)
        sqs_idx_q1 = SingleQubitClifford.find_clifford_index(Y90)
        sq_swap_gates_0_q0 = [(g, "q0") for g in epstein_efficient_decomposition[sqs_idx_q0]]
        sq_swap_gates_0_q1 = [(g, "q1") for g in epstein_efficient_decomposition[sqs_idx_q1]]

        sqs_idx_q0 = SingleQubitClifford.find_clifford_index(Y90)
        sqs_idx_q1 = SingleQubitClifford.find_clifford_index(mY90)
        sq_swap_gates_1_q0 = [(g, "q0") for g in epstein_efficient_decomposition[sqs_idx_q0]]
        sq_swap_gates_1_q1 = [(g, "q1") for g in epstein_efficient_decomposition[sqs_idx_q1]]

        sqs_idx_q1 = SingleQubitClifford.find_clifford_index(Y90)
        sq_swap_gates_2_q0 = [(g, "q0") for g in epstein_efficient_decomposition[0]]
        sq_swap_gates_2_q1 = [(g, "q1") for g in epstein_efficient_decomposition[sqs_idx_q1]]

        gates = (
            C1_q0
            + C1_q1
            + CZ
            + sq_swap_gates_0_q0
            + sq_swap_gates_0_q1
            + CZ
            + sq_swap_gates_1_q0
            + sq_swap_gates_1_q1
            + CZ
            + sq_swap_gates_2_q0
            + sq_swap_gates_2_q1
        )
        return gates


import numpy as np
from qutip import Qobj

def cczs(args:tuple) -> Qobj:
    """
    Unitary of the cczs-gate

    Parameters
    ----------
    args: Tuple
        Input angles for the gate.

    Returns
    -------
    U: Qobj
        The unitary for the cczs-gate.
    """
    theta, phi, gamma = args
    U = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [
                0,
                0,
                0,
                0,
                0,
                -np.exp(-1j * gamma) * np.sin(theta / 2) ** 2 + np.cos(theta / 2) ** 2,
                (1 / 2) * (1 + np.exp(-1j * gamma)) * np.exp(-1j * phi) * np.sin(theta),
                0,
            ],
            [
                0,
                0,
                0,
                0,
                0,
                (1 / 2) * (1 + np.exp(-1j * gamma)) * np.exp(1j * phi) * np.sin(theta),
                -np.exp(-1j * gamma) * np.cos(theta / 2) ** 2 + np.sin(theta / 2) ** 2,
                0,
            ],
            [0, 0, 0, 0, 0, 0, 0, -np.exp(1j * gamma)],
        ],
        dtype="complex",
    )
    U = Qobj(U, dims=[[2] * 3, [2] * 3])
    return U



from chalmers_qubit.devices.sarimner.model import SarimnerModel
from chalmers_qubit.devices.sarimner.compiler import SarimnerCompiler
from chalmers_qubit.devices.sarimner.processor import SarimnerProcessor
from chalmers_qubit.devices.sarimner.noise import DecoherenceNoise, ZZCrossTalk


import warnings
from copy import deepcopy
from typing import Optional
import numpy as np

import qutip
from qutip import propagator, Qobj, QobjEvo
from qutip_qip.circuit import QubitCircuit
from qutip_qip.device import Processor, Model
from qutip_qip.device.processor import _pulse_interpolate
from qutip_qip.compiler import GateCompiler

from chalmers_qubit.devices.sarimner.compiler import SarimnerCompiler


class SarimnerProcessor(Processor):
    """
    Initialize a new SarimnerProcessor instance with a quantum model, an optional compiler, and noise models.

    Parameters
    ----------
    model : Model
        The quantum model that defines the physical properties and capabilities of the processor.
    compiler : GateCompiler, optional
        The compiler used to translate quantum gates into executable operations. If not provided,
        a default compiler specific to the model (SarimnerCompiler) is instantiated and used.
    noise : list, optional
        A list of noise models to be added to the processor. Each element in the list should be compatible
        with the processor's noise handling methods.

    Attributes
    ----------
    model : Model
        The model of the quantum processor, storing physical properties.
    _default_compiler : GateCompiler
        Holds the compiler instance being used, either the provided one or a default SarimnerCompiler.
    native_gates : None
        Initially set to None, to be configured with the gate set natively supported by the processor.
    spline_kind: str
        Type of the coefficient interpolation.
    global_phase : float
        The global phase of the quantum state managed by the processor, initialized to 0.
    """

    def __init__(self,
                 model: Model,
                 compiler: Optional[GateCompiler] = None,
                 noise: Optional[list] = None):

        self.model = model

        if compiler is None:
            self._default_compiler = SarimnerCompiler(model=model)
        else:
            self._default_compiler = compiler

        if noise is not None:
            for elem in noise:
                self.add_noise(elem)

        super(SarimnerProcessor, self).__init__(model=self.model)
        self.native_gates = None
        self.spline_kind = "cubic"
        self.global_phase = 0

    def load_circuit(self, qc: QubitCircuit, schedule_mode: str = "ASAP", compiler: Optional[GateCompiler] = None):
        """
        The default routine of compilation.
        It first calls the :meth:`.transpile` to convert the circuit to
        a suitable format for the hardware model.
        Then it calls the compiler and save the compiled pulses.

        Parameters
        ----------
        qc : :class:`.QubitCircuit`
            Takes the quantum circuit to be implemented.

        schedule_mode: string
            "ASAP" or "ALAP" or None.

        compiler: subclass of :class:`.GateCompiler`
            The used compiler.

        Returns
        -------
        tlist, coeffs: dict of 1D NumPy array
            A dictionary of pulse label and the time sequence and
            compiled pulse coefficients.
        """
        # Choose a compiler and compile the circuit
        if compiler is None and self._default_compiler is not None:
            compiler = self._default_compiler
        if compiler is not None:
            tlist, coeffs = compiler.compile(
                qc.gates, schedule_mode=schedule_mode
            )
        else:
            raise ValueError("No compiler defined.")

        # Update global phase
        self.global_phase = compiler.global_phase

        # Save compiler pulses
        if coeffs is None and tlist is None:
            raise ValueError("The compiled quantum circuit contains no physical pulses.")
        else:
            self.set_coeffs(coeffs)
            self.set_tlist(tlist)
        return tlist, coeffs

    def run_state(
            self,
            init_state=None,
            analytical=False,
            states=None,
            noisy=True,
            solver="mesolve",
            qc=None,
            **kwargs):
        if qc is not None:
            self.load_circuit(qc)
        return super().run_state(init_state, analytical, states, noisy, solver, **kwargs)

    def run_propagator(self, qc: Optional[QubitCircuit] = None, noisy: bool = False, **kwargs):
        """
        Parameters
        ----------
        qc: :class:`qutip.qip.QubitCircuit`, optional
            A quantum circuit. If given, it first calls the ``load_circuit``
            and then calculate the evolution.
        noisy: bool, optional
            If noise are included. Default is False.
        **kwargs
            Keyword arguments for the qutip solver.
        Returns
        -------
        prop: list of Qobj or Qobj
            Returns the propagator(s) calculated at times t.
        """
        if qc is not None:
            self.load_circuit(qc)

        # construct qobjevo
        noisy_qobjevo, sys_c_ops = self.get_qobjevo(noisy=noisy)
        drift_qobjevo = self._get_drift_obj().get_ideal_qobjevo(self.dims)
        H = QobjEvo.__add__(noisy_qobjevo, drift_qobjevo)

        # add collpase operators into kwargs
        if "c_ops" in kwargs:
            if isinstance(kwargs["c_ops"], (Qobj, QobjEvo)):
                kwargs["c_ops"] += [kwargs["c_ops"]] + sys_c_ops
            else:
                kwargs["c_ops"] += sys_c_ops
        else:
            kwargs["c_ops"] = sys_c_ops

        # set time
        if "t" in kwargs:
            t = kwargs["t"]
            del kwargs["t"]
        else:
            tlist = self.get_full_tlist()
            if tlist is None:
                raise ValueError("tlist is None.")
            else:
                t = tlist[-1]

        options = kwargs.get("options", qutip.Options())
        if options.get("max_step", 0.0) == 0.0:
            options["max_step"] = self._get_max_step()
        options["progress_bar"] = False
        kwargs["options"] = options

        # compute the propagator
        prop = propagator(H=H, t=t, **kwargs)

        return prop

    def plot_pulses(
            self,
            title=None,
            figsize=(12, 6),
            dpi=None,
            show_axis=False,
            rescale_pulse_coeffs=True,
            num_steps=1000,
            pulse_labels=None,
            use_control_latex=True,
    ):
        """
        Plot the ideal pulse coefficients.

        Parameters
        ----------
        title: str, optional
            Title for the plot.

        figsize: tuple, optional
            The size of the figure.

        dpi: int, optional
            The dpi of the figure.

        show_axis: bool, optional
            If the axis are shown.

        rescale_pulse_coeffs: bool, optional
            Rescale the hight of each pulses.

        num_steps: int, optional
            Number of time steps in the plot.

        pulse_labels: list of dict, optional
            A map between pulse labels and the labels shown in the y axis.
            E.g. ``[{"sx": "sigmax"}]``.
            Pulses in each dictionary will get a different color.
            If not given and ``use_control_latex==False``,
            the string label defined in each :obj:`.Pulse` is used.

        use_control_latex: bool, optional
            Use labels defined in ``Processor.model.get_control_latex``.

        pulse_labels: list of dict, optional
            A map between pulse labels and the labels shown on the y axis.
            E.g. ``["sx", "sigmax"]``.
            If not given and ``use_control_latex==False``,
            the string label defined in each :obj:`.Pulse` is used.

        use_control_latex: bool, optional
            Use labels defined in ``Processor.model.get_control_latex``.

        Returns
        -------
        fig: matplotlib.figure.Figure
            The `Figure` object for the plot.

        axis: list of ``matplotlib.axes._subplots.AxesSubplot``
            The axes for the plot.

        Notes
        -----
        :meth:.Processor.plot_pulses` only works for array_like coefficients.
        """
        if hasattr(self, "get_operators_labels"):
            warnings.warn(
                "Using the get_operators_labels to provide labels "
                "for plotting is deprecated. "
                "Please use get_control_latex instead."
            )
        import matplotlib.pyplot as plt
        import matplotlib.gridspec as gridspec

        color_list = plt.rcParams["axes.prop_cycle"].by_key()["color"]

        # choose labels
        if pulse_labels is None:
            if use_control_latex and not hasattr(self.model, "get_control_latex"):
                warnings.warn(
                    "No method get_control_latex defined in the model. "
                    "Switch to using the labels defined in each pulse."
                    "Set use_control_latex=False to turn off the warning."
                )
            if use_control_latex:  # use control labels in the model
                control_labels = deepcopy(self.get_control_latex())
                pulse_labels = control_labels
            else:
                pulse_labels = [{pulse.label: pulse.label for pulse in self.pulses}]

        # If it is a nested list instead of a list of dict, we assume that
        if isinstance(pulse_labels[0], list):
            for ind, pulse_group in enumerate(pulse_labels):
                pulse_labels[ind] = {i: latex for i, latex in enumerate(pulse_group)}

        # create a axis for each pulse
        fig = plt.figure(figsize=figsize, dpi=dpi)
        grids = gridspec.GridSpec(sum([len(d) for d in pulse_labels]), 1)
        grids.update(wspace=0.0, hspace=0.0)

        tlist = np.linspace(0.0, self.get_full_tlist()[-1], num_steps)
        dt = tlist[1] - tlist[0]

        # make sure coeffs start and end with zero, for ax.fill
        tlist = np.hstack(([-dt * 1.0e-20], tlist, [tlist[-1] + dt * 1.0e-20]))
        coeffs = []
        for pulse in self.pulses:
            coeffs.append(_pulse_interpolate(pulse, tlist))

        pulse_ind = 0
        axis = []
        for i, label_group in enumerate(pulse_labels):
            for j, (label, latex_str) in enumerate(label_group.items()):
                try:
                    if "cz_real" in label:  # Combine real and imaginaru part of CZ and plot them togehter
                        real_label = label
                        imag_label = real_label.replace("cz_real", "cz_imag")  # Find corresponding cz_imag
                        pulse_real = self.find_pulse(real_label)
                        pulse_imag = self.find_pulse(imag_label)
                        coeff_real = _pulse_interpolate(pulse_real, tlist)
                        coeff_imag = _pulse_interpolate(pulse_imag, tlist)
                        # Combine real and imaginary parts
                        coeff = np.abs(coeff_real + 1j * coeff_imag)
                    elif "cz_imag" in label:
                        coeff = 0
                    else:
                        pulse = self.find_pulse(label)
                        coeff = _pulse_interpolate(pulse, tlist)
                except KeyError:
                    coeff = np.zeros(tlist.shape)
                if ~np.all(coeff == 0):  # only plot pulse if it is non-zero
                    grid = grids[pulse_ind]
                    ax = plt.subplot(grid)
                    axis.append(ax)
                    ax.fill(tlist, coeff, color_list[i], alpha=0.7)
                    ax.plot(tlist, coeff, color_list[i])
                    if rescale_pulse_coeffs:
                        ymax = np.max(np.abs(coeff)) * 1.1
                    else:
                        ymax = np.max(np.abs(coeffs)) * 1.1
                    if ymax != 0.0:
                        ax.set_ylim((-ymax, ymax))

                    # disable frame and ticks
                    if not show_axis:
                        ax.set_xticks([])
                        ax.spines["bottom"].set_visible(False)
                    ax.spines["top"].set_visible(False)
                    ax.spines["right"].set_visible(False)
                    ax.spines["left"].set_visible(False)
                    ax.set_yticks([])
                    ax.set_ylabel(latex_str, rotation=0)
                    pulse_ind += 1
                if i == 0 and j == 0 and title is not None:
                    ax.set_title(title)
        fig.tight_layout()
        return fig, axis



import unittest
from qutip import tensor, basis
from chalmers_qubit.utils.operations import project_on_qubit

class TestQuditProjection(unittest.TestCase):

    def test_project_single_qudit(self):
        # Create a 3-dimensional qudit state |1>
        qudit = basis(3, 1)
        # Expected qubit state |1>
        expected_qubit = basis(2, 1)
        # Project the qudit onto the qubit subspace
        projected_qubit = project_on_qubit(qudit)
        # Assert that the projection is correct
        self.assertTrue(
            projected_qubit == expected_qubit,
            "The projected qudit does not match the expected qubit state.",
        )

    def test_project_two_qudits(self):
        # Create a tensor product of two 3-dimensional qudits |1> ⊗ |1>
        qudit_tensor = tensor(basis(3, 1), basis(3, 1))
        # Expected tensor product of qubits |1> ⊗ |1>
        expected_qubit_tensor = tensor(basis(2, 1), basis(2, 1))
        # Project the tensor product of qudits onto the qubit subspace
        projected_qubit_tensor = project_on_qubit(qudit_tensor)
        # Assert that the projection is correct
        self.assertTrue(
            projected_qubit_tensor == expected_qubit_tensor,
            "The projected qudit tensor does not match the expected qubit tensor state.",
        )


if __name__ == "__main__":
    unittest.main()



import numpy as np
from typing import Optional
from qutip_qip.device.processor import Model
from qutip_qip.compiler import GateCompiler, Instruction
from qutip_qip.operations import Gate

__all__ = ["SarimnerCompiler"]


class SarimnerCompiler(GateCompiler):
    """
    Compiler for :class:`.SarimnerModel`.
    Frequencies are the unit of GHz and times in ns.

    Parameters
    ----------
    model : Model
        Model object of the superconducting qubit device with hardware parameters.
    options : Optional[dict], optional
        A dictionary of compiler options. If not provided, default options will be used.

    Attributes
    ----------
    num_qubits : int
        The number of component systems (qubits).
    params : dict
        A Python dictionary containing the name and value of hardware parameters,
        such as qubit frequency, anharmonicity, etc.
    gate_compiler : dict
        A Python dictionary in the form of {gate_name: decompose_function}.
        It stores the decomposition scheme for each supported gate.
    phase : list
        List of values indicating how much we have virtually rotated the Bloch sphere of each qubit.
    global_phase : float
        The global phase of the quantum state.
    dt : float
        Time step in nanoseconds.
    two_qubit_gate_options : dict
        Options specific to two-qubit gates, including buffer_time and rise_fall_time.
    single_qubit_gate_options : dict
        Options specific to single-qubit gates, including type, gate_time, pi-pulse amplitude, and std.

    Notes
    -----
    The compiler supports various gates including RZ, RX, RY, X, H, CZ, ISWAP, CCZS, IDLE, and GLOBALPHASE.
    Default options are provided for dt, two-qubit gates, and single-qubit gates, which can be overridden
    by passing custom options during initialization.
    """

    def __init__(self, model: Model, options: Optional[dict] = None):
        self.num_qubits = model.num_qubits
        self.params = model.params

        super().__init__(num_qubits=self.num_qubits, params=self.params)
        self.gate_compiler.update(
            {
                "RZ": self.rz_compiler,
                "RX": self.rx_compiler,
                "RY": self.ry_compiler,
                "X": self.x_compiler,
                "H": self.h_compiler,
                "CZ": self.cz_compiler,
                "ISWAP": self.iswap_compiler,
                "CCZS": self.cczs_compiler,
                "IDLE": self.idle_compiler,
                "GLOBALPHASE": self.globalphase_compiler,
            }
        )

        # Parse options
        default_options = {
            "dt": 0.1,  # in (ns)
            "two_qubit_gate": {
                "buffer_time": 1,  # in (ns)
                "rise_fall_time": 1,  # in (ns)
            },
            "single_qubit_gate": {
                "type": "gaussian",
                "gate_time": 50,  # in (ns)
                "pi_pulse_amplitude": 0.12533148558448476,
                "std": 5,  # in (ns)
            },
        }

        if options is None:
            options = {}

        self.options = self._merge_options(default_options, options)

        # Set attributes based on parsed options
        self.dt = self.options["dt"]  # time step in (ns)
        self.two_qubit_gate_options = self.options["two_qubit_gate"]
        self.single_qubit_gate_options = self.options["single_qubit_gate"]
        # initialise the phase for all the qubits to 0.
        self.phase = [0] * self.num_qubits
        # initialise the global phase to 0.
        self.global_phase = 0

    def _merge_options(self, default_options, user_options):
        """Recursively merge user options into default options."""
        for key, value in user_options.items():
            if isinstance(value, dict) and key in default_options:
                default_options[key] = self._merge_options(default_options[key], value)
            else:
                default_options[key] = value
        return default_options

    @staticmethod
    def _envelope_function(t, t_gate, t_rise_fall, t_buffer):
        # Total gate time including rise and fall times
        total_time = t_gate + 2 * t_rise_fall + 2 * t_buffer
        if t_rise_fall > 0:
            # Calculate rise window
            rise_window = np.sin(np.pi * (t - t_buffer) / (2 * t_rise_fall)) ** 2
            # Calculate fall window
            fall_window = np.cos(np.pi * (t - (total_time - t_rise_fall - t_buffer)) / (2 * t_rise_fall)) ** 2
        else:
            rise_window = 0
            fall_window = 0

        # Initialize array with zeros
        arr = np.zeros_like(t)

        # Set rise window values
        arr = np.where((t >= t_buffer) & (t < (t_buffer + t_rise_fall)), rise_window, arr)
        # Set constant window values
        arr = np.where((t >= (t_buffer + t_rise_fall)) & (t < (total_time - t_rise_fall - t_buffer)), 1, arr)
        # Set fall window values
        arr = np.where((t >= (total_time - t_rise_fall - t_buffer)) & (t < (total_time - t_buffer)), fall_window, arr)
        return arr

    @staticmethod
    def _gaussian_drag(t: np.ndarray, op: str, args: dict) -> np.ndarray:
        # Amplitude, needs to be optimized to get perfect pi-pulse or pi/2-pulse
        amp = args['amp']
        # DRAG-parameter, needs to be optimized to get no phase errors for a pi/2-pulse
        qscale = args['qscale']
        drive_freq = args['freq']
        phase = args['phase']
        L = args['gatetime']
        sigma = args['sigma']
        Omega_x = amp * np.exp(-pow(t - 0.5 * L, 2) / (2 * pow(sigma, 2)))
        Omega_y = qscale * (t - 0.5 * L) / pow(sigma, 2) * Omega_x

        if op == "x":
            coeff = Omega_x * np.cos(drive_freq * t + phase) \
                    + Omega_y * np.sin(drive_freq * t + phase)
        elif op == "y":
            coeff = Omega_x * np.sin(drive_freq * t + phase) \
                    - Omega_y * np.cos(drive_freq * t + phase)
        return coeff

    def _rotation_gate(self, gate, phase):
        """
        Compiler for the rotational gate that lies along the equator gate

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        Returns
        -------
        A list of :obj:`.Instruction`, including the compiled pulse
        information for this gate.
        """

        # target qubit
        target = gate.targets[0]

        # Use the parsed options for single qubit gates
        sigma = self.single_qubit_gate_options["std"]
        amp = self.single_qubit_gate_options["pi_pulse_amplitude"] * gate.arg_value / np.pi
        t_gate = self.single_qubit_gate_options["gate_time"]

        # parameters
        alpha = self.params["transmons"][target]["anharmonicity"]
        omega_qubit = self.params["transmons"][target]["frequency"]
        rotating_freq = omega_qubit
        omega_drive = rotating_freq - omega_qubit

        # arguments
        args = {
            "amp": amp,
            "qscale": -0.5 / alpha,
            "phase": phase,
            "freq": omega_drive,
            "gatetime": t_gate,
            "sigma": sigma,
        }

        # time list
        tlist = np.arange(0, t_gate + self.dt, self.dt)

        # set start and end to zero
        coeff_x = self._gaussian_drag(tlist, "x", args)
        coeff_y = self._gaussian_drag(tlist, "y", args)

        pulse_info = [
            # (control label, coeff)
            ("x" + str(target), coeff_x),
            ("y" + str(target), coeff_y),
        ]

        return [Instruction(gate, tlist, pulse_info, t_gate)]

    def rx_compiler(self, gate, args):
        """
        Compiler for the RX gate

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        Returns
        -------
        A list of :obj:`.Instruction`, including the compiled pulse
        information for this gate.
        """

        # target qubit
        target = gate.targets[0]
        # phase
        phase = self.phase[target]
        return self._rotation_gate(gate=gate, phase=phase)

    def cz_compiler(self, gate, args):
        """
        Compiler for CZ gate.

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        Returns
        -------
        A list of :obj:`.Instruction`, including the compiled pulse
        information for this gate.
        """
        q1 = gate.controls[0]
        q2 = gate.targets[0]

        # Coupling strength
        g = self.params["couplings"][q1, q2]

        # Get frequencies
        alpha1 = self.params["transmons"][q1]["anharmonicity"]
        delta = alpha1

        # Total time of gate
        t_gate = np.pi / (np.sqrt(2) * abs(g))
        # Rise and fall time of envelope
        rise_fall_time = self.two_qubit_gate_options["rise_fall_time"]
        # Buffer time
        t_buffer = self.two_qubit_gate_options["buffer_time"]
        # Total time
        t_total = t_gate + 2 * rise_fall_time + 2 * t_buffer
        # Time list
        tlist = np.arange(0, t_total + self.dt, self.dt)

        coeff = self._envelope_function(tlist, t_gate, rise_fall_time, t_buffer)

        pulse_info = [
            ("cz_real" + str(q1) + str(q2), coeff * np.cos(delta * tlist)),
            ("cz_imag" + str(q1) + str(q2), coeff * np.sin(delta * tlist)),
        ]

        return [Instruction(gate, tlist, pulse_info, t_total)]

    def rz_compiler(self, gate, args):
        """
        Compiler for the Virtual-RZ gate

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        """
        q = gate.targets[0]  # target qubit
        # This corresponds to clockwise rotation of the Bloch-sphere around the Z-axis.
        self.phase[q] -= gate.arg_value

    def ry_compiler(self, gate, args):
        """
        Compiler for the RY gate

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        Returns
        -------
        A list of :obj:`.Instruction`, including the compiled pulse
        information for this gate.
        """
        # target qubit
        target = gate.targets[0]
        # phase
        phase = self.phase[target] + np.pi / 2
        return self._rotation_gate(gate=gate, phase=phase)

    def x_compiler(self, gate, args):
        """
        Compiler for the Hadamard gate

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        """
        q = gate.targets[0]  # target qubit
        rx_gate = self.gate_compiler["RX"](
            Gate("RX", [q], None, arg_value=np.pi),
            None)
        self.gate_compiler["GLOBALPHASE"](
            Gate("GLOBALPHASE", [q], None, arg_value=np.pi / 2),
            None)
        # we only return the physical gate
        return rx_gate

    def h_compiler(self, gate, args):
        """
        Compiler for the Hadamard gate

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        """
        q = gate.targets[0]  # target qubit

        self.gate_compiler["RZ"](
            Gate("RZ", [q], None, arg_value=np.pi),
            None)
        ry_gate = self.gate_compiler["RY"](
            Gate("RY", [q], None, arg_value=np.pi / 2),
            None)
        self.gate_compiler["GLOBALPHASE"](
            Gate("GLOBALPHASE", [q], None, arg_value=np.pi / 2),
            None)
        # we only return the physical gate
        return ry_gate

    def iswap_compiler(self, gate, args):
        """
        Compiler for ISWAP gate.

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        Returns
        -------
        A list of :obj:`.Instruction`, including the compiled pulse
        information for this gate.
        """
        q1 = gate.controls[0]
        q2 = gate.targets[0]

        # Coupling strength
        g = self.params["couplings"][q1, q2]

        # Total time of gate
        t_gate = np.pi / (2 * abs(g))
        # Rise and fall time of envelope
        rise_fall_time = self.two_qubit_gate_options["rise_fall_time"]
        # Buffer time
        t_buffer = self.two_qubit_gate_options["buffer_time"]
        # Total time
        t_total = t_gate + 2 * rise_fall_time + 2 * t_buffer
        # Time list
        tlist = np.arange(0, t_total + self.dt, self.dt)

        coeff = self._envelope_function(tlist, t_gate, rise_fall_time, t_buffer)

        pulse_info = [("iswap" + str(q1) + str(q2), coeff)]
        # ADD VIRTUAL-Z GATES TO CORRECT THE PHASE
        self.phase[q1] -= np.pi
        self.phase[q2] -= np.pi
        return [Instruction(gate, tlist, pulse_info, t_total)]

    def cczs_compiler(self, gate, args):
        """
        Compiler for CCZS gate.

        Parameters
        ----------
        gate : :obj:`.Gate`:
            The quantum gate to be compiled.
        args : dict
            The compilation configuration defined in the attributes
            :obj:`.GateCompiler.args` or given as a parameter in
            :obj:`.GateCompiler.compile`.

        Returns
        -------
        A list of :obj:`.Instruction`, including the compiled pulse
        information for this gate.
        """
        q1 = gate.targets[0]  # this is the control qubit
        q2 = gate.targets[1]
        q3 = gate.targets[2]
        theta, phi, gamma = gate.arg_value

        # define parameters
        alpha1 = self.params["transmons"][q1]["anharmonicity"]

        g1 = self.params["couplings"][q1, q2]
        g2 = self.params["couplings"][q1, q3]
        g = np.sqrt(np.abs(g1) ** 2 + np.abs(g2) ** 2)

        # Total time of gate
        t_gate = np.pi / (np.sqrt(2) * g)
        # Rise and fall time of envelope
        rise_fall_time = self.two_qubit_gate_options["rise_fall_time"]
        # Buffer time
        t_buffer = self.two_qubit_gate_options["buffer_time"]
        # Total time
        t_total = t_gate + 2 * rise_fall_time + 2 * t_buffer
        # Time list
        tlist = np.arange(0, t_total + self.dt, self.dt)

        coeff = self._envelope_function(tlist, t_gate, rise_fall_time, t_buffer)

        pulse_info = [
            ("cz_real" + str(q1) + str(q2), coeff * np.cos(alpha1 * tlist)),
            ("cz_imag" + str(q1) + str(q2), coeff * np.sin(alpha1 * tlist)),
            ("cz_real" + str(q1) + str(q3), coeff * np.cos(alpha1 * tlist)),
            ("cz_imag" + str(q1) + str(q3), coeff * np.sin(alpha1 * tlist)),
        ]

        return [Instruction(gate, tlist, pulse_info, t_total)]

    def globalphase_compiler(self, gate, args):
        """
        Compiler for the GLOBALPHASE gate
        """
        self.global_phase += gate.arg_value
        return [Instruction(gate, self.global_phase, [])]

    def idle_compiler(self, gate, args):
        """
        Compiler for the IDLE gate
        """
        # target qubit
        q = gate.targets[0]
        idle_time = gate.arg_value

        if idle_time < 0:
            ValueError("Error: Idle time cannot be negative.")
        elif (0 <= idle_time < self.dt):
            # Skip the gate if idle-time is less than dt.
            pass
        else:
            # Take one time step for every nano second
            n_steps = int(np.ceil(idle_time))
            # Make sure that we at least take two steps
            if n_steps < 2:
                n_steps = 2
            tlist = np.linspace(0, idle_time, n_steps)
            coeff = np.zeros(n_steps)
            # Create a pulse with zero amplitude
            pulse_info = [
                ("x" + str(q), coeff),
                ("y" + str(q), coeff)
            ]
            return [Instruction(gate, tlist, pulse_info, idle_time)]



import unittest
import numpy as np
from qutip import basis, tensor, fidelity, average_gate_fidelity
from qutip_qip.circuit import QubitCircuit
from qutip_qip.operations import rz
from chalmers_qubit.devices.sarimner import SarimnerModel, SarimnerProcessor, SarimnerCompiler
from chalmers_qubit.utils.operations import project_on_qubit
from chalmers_qubit.utils.gates import cczs

import warnings
warnings.simplefilter(action="ignore", category=FutureWarning)

class TestSingleQubitGates(unittest.TestCase):
    def setUp(self) -> None:
        self.num_qubits = 1
        transmon_dict = {0: {"frequency": 5.0, "anharmonicity": 0.3}}
        self.model = SarimnerModel(transmon_dict=transmon_dict)
        self.processor = SarimnerProcessor(model=self.model)
        return super().setUp()

    def tearDown(self) -> None:
        return super().tearDown()

    def test_global_phase_gate(self):
        qc = QubitCircuit(self.num_qubits)
        arg_value = np.pi
        qc.add_gate("RX", targets=0, arg_value=np.pi/2)
        qc.add_gate("GLOBALPHASE", targets=0, arg_value=arg_value)
        qc.add_gate("RZ", targets=0, arg_value=np.pi/2)
        self.processor.load_circuit(qc)
        global_phase = self.processor.global_phase
        self.assertEqual(global_phase, arg_value)

    def test_rx_gate(self):
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("RX", targets=0, arg_value=np.pi/2)
        # Compute the ideal propagator
        U = qc.compute_unitary()
        # Compute the propagator
        prop = self.processor.run_propagator(qc=qc)
        final_prop = project_on_qubit(prop)
        # Compute the average gate fidelity
        f = average_gate_fidelity(U, final_prop)
        # We want 99.99% Average Gate Fidelity
        self.assertAlmostEqual(1, f, places=4, msg="Precision of RX-gate failed.")

    def test_ry_gate(self):
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("RY", targets=0, arg_value=np.pi/2)
        # Compute the ideal propagator
        U = qc.compute_unitary()
        # Compute the propagator
        prop = self.processor.run_propagator(qc=qc)
        final_prop = project_on_qubit(prop)
        # Compute the average gate fidelity
        f = average_gate_fidelity(U, final_prop)
        # We want 99.99% Average Gate Fidelity
        self.assertAlmostEqual(1, f, places=4, msg="Precision of RY-gate failed.")

    def test_rz_gate(self):
        qc = QubitCircuit(self.num_qubits)
        # Since the RZ-gate is virtual we need at least
        # one physical gate to test it.
        qc.add_gate("RZ", targets=0, arg_value=np.pi / 2)
        qc.add_gate("RX", targets=0, arg_value=np.pi / 2)
        # Prepare "plus"-state as initial state
        init_state = (basis(3, 0) + basis(3, 1)) / np.sqrt(2)
        result = self.processor.run_state(init_state, qc=qc)
        final_state = project_on_qubit(result.states[-1])
        ideal_state = basis(2, 0)
        f = fidelity(final_state, ideal_state)
        # We want 99.99% average state fidelity
        self.assertAlmostEqual(1, f, places=4, msg="Precision of RZ-gate failed.")

    def test_no_physical_pulse(self):
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("RZ", targets=0, arg_value=np.pi / 2)
        init_state = (basis(3, 0) + basis(3, 1)) / np.sqrt(2)
        # Raise value error if quantum circuit contain no physical pulses.
        with self.assertRaises(ValueError):
            self.processor.run_state(init_state, qc=qc)

    def test_x_gate(self):
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("X", targets=0)
        # Compute the ideal propagator
        U = qc.compute_unitary()
        # Compute the propagator
        prop = self.processor.run_propagator(qc=qc)
        final_prop = project_on_qubit(prop)
        # Compute the average gate fidelity
        f = average_gate_fidelity(U, final_prop)
        # We want 99.99% Average Gate Fidelity
        self.assertAlmostEqual(1, f, places=4, msg="Precision of X-gate failed.")

    def test_hadamard_gate(self):
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("H", targets=0)
        # Compute the ideal propagator
        U = qc.compute_unitary()
        # Compute the propagator
        prop = self.processor.run_propagator(qc=qc)
        # Since we have Virtually rotated the Bloch sphere 180-degrees
        # we need to rotate it back to get the correct propagator.
        final_prop = rz(np.pi) * project_on_qubit(prop)
        # Compute the average gate fidelity
        f = average_gate_fidelity(U, final_prop)
        # We want 99.99% Average Gate Fidelity
        self.assertAlmostEqual(1, f, places=4, msg="Precision of H-gate failed")

    def test_idling_gate(self):
        t_total = 1e3 # time in (ns)
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("IDLE", targets=0, arg_value=t_total)
        # Simulate
        init_state = (basis(3, 0) + basis(3, 1)).unit()
        result = self.processor.run_state(init_state, qc=qc)
        final_state = result.states[-1]
        # Compute the state fidelity
        f = fidelity(final_state, init_state)
        self.assertAlmostEqual(1, f, places=7)

    def test_single_qubit_circuit(self):
        # Create a somewhat random qubit circuit
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("RX", targets=0, arg_value=np.pi/2)
        qc.add_gate("RZ", targets=0, arg_value=np.pi/2)
        qc.add_gate("RY", targets=0, arg_value=np.pi/2)
        qc.add_gate("RZ", targets=0, arg_value=-np.pi/2)
        qc.add_gate("RX", targets=0, arg_value=np.pi)
        # Simulate
        init_state = basis(3, 0)
        result = self.processor.run_state(init_state, qc=qc)
        final_state = result.states[-1]
        # Project final state to the qubit subspace
        qubit_state = project_on_qubit(final_state)
        qubit_init_state = project_on_qubit(init_state)
        # Get the ideal target state
        target_state = qc.compute_unitary() * qubit_init_state
        # Compute the fidelity
        f = fidelity(target_state, qubit_state)
        self.assertAlmostEqual(1, f, places=3)


class TestTwoQubitGates(unittest.TestCase):
    def setUp(self) -> None:
        self.num_qubits = 2
        transmon_dict = {
            0: {"frequency": 5.0, "anharmonicity": 0.3},
            1: {"frequency": 5.4, "anharmonicity": 0.3},
        }
        coupling_dict = {(0,1): 1e-3}
        options = {
            "dt": 0.1,  # time step in (ns)
            "two_qubit_gate": {
                "buffer_time": 0,
                "rise_fall_time": 0,
            },
        }
        self.model = SarimnerModel(transmon_dict=transmon_dict, 
                                   coupling_dict=coupling_dict)
        compiler = SarimnerCompiler(model=self.model, options=options)
        self.processor = SarimnerProcessor(model=self.model, compiler=compiler)
        return super().setUp()

    def tearDown(self) -> None:
        return super().tearDown()

    def test_cz_gate(self):
        # Create a circuit
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("CZ", controls=0, targets=1)
        # Compute the ideal propagator
        U = qc.compute_unitary()
        # Compute the propagator
        prop = self.processor.run_propagator(qc=qc, options={"nsteps": 1e5})
        final_prop = project_on_qubit(prop)
        # Compute the average gate fidelity
        f = average_gate_fidelity(U, final_prop)
        # We want 99.99% Average Gate Fidelity
        self.assertAlmostEqual(1, f, places=4)

    def test_iswap_gate(self):
        # Create a circuit
        qc = QubitCircuit(self.num_qubits)
        qc.add_gate("ISWAP", controls=0, targets=1)
        # Compute the ideal propagator
        U = qc.compute_unitary()
        # Compute the propagator
        prop = self.processor.run_propagator(qc=qc, options={"nsteps": 1e5})
        # Need to append the "Virtual RZ-gates"
        final_prop = tensor(rz(np.pi), rz(np.pi)) * project_on_qubit(prop)
        # Compute the average gate fidelity
        f = average_gate_fidelity(U, final_prop)
        self.assertAlmostEqual(1, f, places=4)


class TestThreeQubitGates(unittest.TestCase):
    def setUp(self) -> None:
        self.num_qubits = 3
        transmon_dict = {
            0: {"frequency": 5.3, "anharmonicity": 0.3},
            1: {"frequency": 5.1, "anharmonicity": 0.3},
            2: {"frequency": 5.2, "anharmonicity": 0.3},
        }
        coupling_dict = {(0, 1): 1e-3,
                         (0, 2): 1e-3}
        # Compiler options
        options = {
            "dt": 0.1, # time step in (ns)
            "two_qubit_gate": {
                "buffer_time": 0,
                "rise_fall_time": 0,
            },
        }
        self.model = SarimnerModel(transmon_dict=transmon_dict,
                                   coupling_dict=coupling_dict)
        compiler = SarimnerCompiler(model=self.model, options=options)
        self.processor = SarimnerProcessor(model=self.model, compiler=compiler)
        return super().setUp()

    def tearDown(self) -> None:
        return super().tearDown()

    def test_cczs_gate(self):
        # Create a circuit
        qc = QubitCircuit(self.num_qubits)
        qc.user_gates = {"CCZS": cczs}
        qc.add_gate("CCZS", targets=[0,1,2], arg_value=[np.pi/2,np.pi,0])
        # Compute the ideal propagator
        U = qc.compute_unitary()
        # Compute the propagator
        prop = self.processor.run_propagator(qc=qc, options={"nsteps": 1e5})
        final_prop = project_on_qubit(prop)
        # Compute the average gate fidelity
        f = average_gate_fidelity(U, final_prop)
        # We want 99.99% Average Gate Fidelity
        self.assertAlmostEqual(1, f, places=4)

if __name__ == "__main__":
    unittest.main()



import numpy as np
from qutip import Qobj


def project_on_qubit(rho: Qobj) -> Qobj:
    """
    Project a given quantum object (density matrix or state vector) onto the computational basis of qubits.

    Parameters
    ----------
    rho : Qobj
        The input quantum object, either a density matrix ('oper' type) or a state vector.

    Returns
    -------
    qubit_state : Qobj
        The projected quantum object onto the computational basis (0 and 1 states) of qubits.
    """
    if rho.dims is None:
        raise ValueError("The 'dims' attribute of the input Qobj is None. Please ensure it is correctly initialized.")

    dims = rho.dims
    num_qubits = len(dims[0])
    N = dims[0][0]
    # Generate all computational basis states in base N
    base_N = [np.base_repr(i, base=N) for i in range(N**num_qubits)]
    # Generate all computational basis states in base 2
    base_2 = [np.base_repr(i, base=2) for i in range(2**num_qubits)]
    # Find indices of computational basis states in base N list
    l = [base_N.index(i) for i in base_2]
    if rho.type == 'oper':
        # If rho is a density matrix, project it onto the computational basis
        qubit_state = Qobj([[rho.full()[j, i] for i in l] for j in l], dims=[
                           [2] * num_qubits, [2] * num_qubits])
    else:
        # If rho is a state vector, project it onto the computational basis
        qubit_state = Qobj([rho.full()[j] for j in l], dims=[
                           [2] * num_qubits, [1] * num_qubits])
    return qubit_state


import numpy as np
from qutip import Qobj
from itertools import product
from functools import reduce

# Predefined Pauli matrices
SIGMAX = np.array([[0, 1], [1, 0]])
SIGMAY = np.array([[0, -1j], [1j, 0]])
SIGMAZ = np.array([[1, 0], [0, -1]])
PAULI_MATRICES = [np.eye(2), SIGMAX, SIGMAY, SIGMAZ]

# TODO: Write Test cases for the functions in this module
def pauli_basis(n_qubits: int) -> list:
    """Generate n-qubit Pauli basis using Kronecker product."""
    basis = []
    for indices in product(range(4), repeat=n_qubits):
        ops = [PAULI_MATRICES[i] for i in indices]
        kron_prod = reduce(np.kron, ops)
        basis.append(kron_prod)
    return basis

# TODO: Write Test cases for the functions in this module
def unitary_to_ptm(U: Qobj) -> np.ndarray:
    """Convert a unitary Qobj to its Pauli Transfer Matrix (PTM)."""
    U_np = U.full()
    n_qubits = int(np.log2(U.shape[0]))
    d = 2 ** n_qubits
    pauli_ops = pauli_basis(n_qubits)

    ptm = np.zeros((d**2, d**2), dtype=float)
    for i, Pi in enumerate(pauli_ops):
        for j, Pj in enumerate(pauli_ops):
            q_map = U_np @ Pj @ U_np.conj().T
            ptm[i, j] = np.trace(Pi @ q_map).real / d

    return ptm


# Modifications copyright (c) 2025 Pontus Vikstål
#
# MIT License
# 
# Copyright (c) 2016 DiCarlo lab-QuTech-Delft University of Technology
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from typing import Optional, Union, Literal, Dict, List
from chalmers_qubit.utils.randomized_benchmarking.clifford_group import SingleQubitClifford, TwoQubitClifford, Clifford
from chalmers_qubit.utils.transformation import unitary_to_ptm
from qutip import Qobj
from qutip_qip.circuit import QubitCircuit
from qutip_qip.operations import RX, RY, CZ, Gate

__all__ = [
    "RandomizedBenchmarking",
]

class RandomizedBenchmarking:
    """
    Implementation of Clifford-based Randomized Benchmarking (RB) for quantum gates.

    This class supports standard and interleaved randomized benchmarking for one and 
    two-qubit Clifford groups. It generates random Clifford sequences and converts them 
    to quantum circuits with physical gates.

    Attributes
    ----------
    clifford_group : int
        Specifies which Clifford group to use (1 for single-qubit, 2 for two-qubit).
    CliffordClass : Type
        The class representing Clifford operations for the chosen group.
    """

    def __init__(self, clifford_group: Literal[1, 2] = 1) -> None:
        """
        Initialize the RandomizedBenchmarking class.

        Parameters
        ----------
        clifford_group : Literal[1, 2], optional
            Specifies which Clifford group to use.
            1 for single-qubit (24 elements), 2 for two-qubit (11,520 elements).

        Raises
        ------
        NotImplementedError
            If an unsupported Clifford group is specified.
        """
        clifford_classes = {
            1: SingleQubitClifford,
            2: TwoQubitClifford,
        }
        if clifford_group not in clifford_classes:
            raise NotImplementedError("Only one- and two-qubit Clifford groups (1 or 2) are supported.")

        self.CliffordClass = clifford_classes[clifford_group]
        self.clifford_group = clifford_group

    def __repr__(self) -> str:
        """
        Return a string representation of the RandomizedBenchmarking instance.

        Returns
        -------
        str
            String representation of the instance.
        """
        return f"RandomizedBenchmarking(clifford_group={self.clifford_group})"

    def _calculate_net_clifford(
            self,    
            clifford_indices: np.ndarray,
        ) -> "Clifford":
        """
        Calculate the net Clifford from a list of Clifford indices.

        Parameters
        ----------
        clifford_indices : np.ndarray
            Array of integers specifying the Cliffords.

        Returns
        -------
        Clifford
            A `Clifford` object containing the net Clifford. The Clifford index is contained in the `Clifford.idx` attribute.

        Notes
        -----
        The order corresponds to the order in a pulse sequence but is
        the reverse of what it would be in a chained dot product.
        """
        net_clifford = self.CliffordClass(0) # assumes element 0 is the Identity
        for idx in clifford_indices:
            clifford = self.CliffordClass(idx)
            net_clifford = clifford * net_clifford
        return net_clifford

    def _add_interleaved_clifford_idx(
            self,
            clifford_sequence: np.ndarray, 
            interleaved_clifford_idx: int
        ) -> np.ndarray:
        """
        Add the interleaved Clifford gate index to the sequence.

        Parameters
        ----------
        clifford_sequence : np.ndarray
            Array of Clifford indices.
        interleaved_clifford_idx : int
            Interleaved Clifford index.

        Returns
        -------
        np.ndarray
            Clifford sequence with interleaved Clifford.
        """
        interleaved_sequence = np.empty(clifford_sequence.size * 2, dtype=int)
        interleaved_sequence[0::2] = clifford_sequence
        interleaved_sequence[1::2] = interleaved_clifford_idx
        return interleaved_sequence

    def _add_inverse_clifford_idx(
        self,
        clifford_sequence: np.ndarray,
    ) -> np.ndarray:
        """
        Find and add the inverse of the total sequence to the end of the sequence.

        Parameters
        ----------
        clifford_sequence : np.ndarray
            Array of Clifford indices.

        Returns
        -------
        np.ndarray
            Array with appended inverse Clifford index.
        """
        net_clifford = self._calculate_net_clifford(clifford_sequence)
        inverse_clifford = net_clifford.get_inverse() 
        return np.append(clifford_sequence, inverse_clifford.idx)
    
    def _get_interleaved_clifford_idx(
        self,
        interleaved_clifford_gate: Optional[Union[Gate, QubitCircuit]],
    ) -> int:
        """
        Get the index of the interleaved Clifford gate.

        Parameters
        ----------
        interleaved_clifford_gate : Optional[Union[Gate, QubitCircuit]]
            The interleaved Clifford gate.

        Returns
        -------
        int
            The index of the interleaved Clifford gate.

        Raises
        ------
        ValueError
            If the input is not a Gate or QubitCircuit, or if the unitary dimension does not match the Clifford group.
        """
        # Get the unitary representation of the interleaved Clifford gate
        if isinstance(interleaved_clifford_gate, QubitCircuit):
            interleaved_clifford_unitary = interleaved_clifford_gate.compute_unitary()
        elif isinstance(interleaved_clifford_gate, Gate):
            interleaved_clifford_unitary = interleaved_clifford_gate.get_compact_qobj()
        else:
            raise ValueError(f"interleaved_clifford must be {QubitCircuit} or {Gate}.")
        
        # Check that the unitary dimension matches the clifford group
        if interleaved_clifford_unitary.shape[0] != 2**self.clifford_group:
            raise ValueError(f"Interleaved Clifford unitary dimension {interleaved_clifford_unitary.shape[0]} does not match the dimension of clifford group {self.clifford_group}.")
        
        # Convert the unitary to a PTM and find the corresponding Clifford index
        ptm = unitary_to_ptm(interleaved_clifford_unitary)
        interleaved_clifford_idx = self.CliffordClass.find_clifford_index(ptm)
        return interleaved_clifford_idx
    
    def _add_interleaved_gate(self, circuit: QubitCircuit, gate: Union[Gate, QubitCircuit]):
        """
        Add an interleaved gate or circuit to the given QubitCircuit.

        Parameters
        ----------
        circuit : QubitCircuit
            The circuit to which the gate will be added.
        gate : Union[Gate, QubitCircuit]
            The gate or circuit to add.

        Raises
        ------
        ValueError
            If the input is not a Gate or QubitCircuit.
        """
        if isinstance(gate, QubitCircuit):
            circuit.add_circuit(gate)
        elif isinstance(gate, Gate):
            circuit.add_gate(gate)
        else:
            raise ValueError(f"interleaved_clifford_gate must be a {Gate} or {QubitCircuit}.")
    
    def _gate_decomposition(
            self,
            clifford_indices: np.ndarray,
            interleaved_clifford_gate: Optional[Union[Gate, QubitCircuit]] = None,
    ) -> QubitCircuit:
        """
        Decompose a sequence of Clifford indices into physical gates.

        Parameters
        ----------
        clifford_indices : np.ndarray
            Array of Clifford indices.
        interleaved_clifford_gate : Optional[Union[Gate, QubitCircuit]], optional
            Interleaved Clifford gate or circuit.

        Returns
        -------
        QubitCircuit
            The physical circuit with gates.

        Raises
        ------
        ValueError
            If a Clifford gate has no decomposition.
        """
        # Create a mapping of qubit indices to qubit names
        qubit_map = {f"q{idx}": idx for idx in range(self.clifford_group)}
        operation_map = {
            "X180": lambda q: RX(targets=qubit_map[q], arg_value=np.pi),
            "X90": lambda q: RX(targets=qubit_map[q], arg_value=np.pi/2),
            "Y180": lambda q: RY(targets=qubit_map[q], arg_value=np.pi),
            "Y90": lambda q: RY(targets=qubit_map[q], arg_value=np.pi/2),
            "mX90": lambda q: RX(targets=qubit_map[q], arg_value=-np.pi/2),
            "mY90": lambda q: RY(targets=qubit_map[q], arg_value=-np.pi/2),
            "CZ": lambda q: CZ(controls=qubit_map[q[0]], targets=qubit_map[q[1]]),
        }

        # Initialize the circuit
        circuit = QubitCircuit(self.clifford_group)

        for i, clifford_idx in enumerate(clifford_indices):
            if interleaved_clifford_gate is not None and (i+1) % 2 == 0:
                self._add_interleaved_gate(circuit, interleaved_clifford_gate)
                continue
 
            cl_decomp = self.CliffordClass(clifford_idx).gate_decomposition
            if cl_decomp is None:
                raise ValueError(f"Clifford gate {clifford_idx} has no decomposition.")
            for gate, q in cl_decomp:
                if gate == "I":
                    continue
                operation = operation_map[gate](q)
                circuit.add_gate(operation)
        return circuit

    def _randomized_benchmarking_sequence(
        self,
        number_of_cliffords: int,
        apply_inverse: bool = True,
        interleaved_clifford_idx: Optional[int] = None,
        seed: Optional[int] = None,
    ) -> np.ndarray:
        """
        Generate a randomized benchmarking sequence using the one- or two-qubit Clifford group.

        Parameters
        ----------
        number_of_cliffords : int
            Number of Clifford gates in the sequence (excluding the optional inverse).
        apply_inverse : bool, optional
            Whether to append the recovery Clifford that inverts the total sequence.
        interleaved_clifford_idx : Optional[int], optional
            Optional ID for interleaving a specific Clifford gate.
        seed : Optional[int], optional
            Optional seed for reproducibility.

        Returns
        -------
        np.ndarray
            Array of Clifford indices representing the randomized benchmarking sequence.

        Raises
        ------
        ValueError
            If `number_of_cliffords` is negative.
        """
        if number_of_cliffords < 1:
            raise ValueError("Number of Cliffords must at least be 1.")
        
        group_size = self.CliffordClass.GROUP_SIZE
        rng = np.random.default_rng(seed)
        clifford_indices = rng.integers(low=0, high=group_size, size=number_of_cliffords)

        if interleaved_clifford_idx is not None:
            clifford_indices = self._add_interleaved_clifford_idx(
                clifford_sequence=clifford_indices, 
                interleaved_clifford_idx=interleaved_clifford_idx
            )

        if apply_inverse:
            clifford_indices = self._add_inverse_clifford_idx(clifford_sequence=clifford_indices)

        return clifford_indices

    def randomized_benchmarking_circuit(
        self,
        number_of_cliffords: int,
        apply_inverse: bool = True,
        interleaved_clifford_gate: Optional[Union[Gate, QubitCircuit]] = None,
        seed: Optional[int] = None,
    ) -> QubitCircuit:
        """
        Generate a randomized benchmarking circuit from a sequence of Clifford indices.

        Parameters
        ----------
        number_of_cliffords : int
            Number of Cliffords in the sequence.
        apply_inverse : bool, optional
            Whether to append the recovery Clifford that inverts the total sequence.
        interleaved_clifford_gate : Optional[Union[Gate, QubitCircuit]], optional
            Optional interleaved Clifford gate or circuit.
        seed : Optional[int], optional
            Optional seed for reproducibility.

        Returns
        -------
        QubitCircuit
            The randomized benchmarking circuit.
        """
        if interleaved_clifford_gate is not None:
            interleaved_clifford_idx = self._get_interleaved_clifford_idx(interleaved_clifford_gate)
        else:
            interleaved_clifford_idx = None

        clifford_indices = self._randomized_benchmarking_sequence(
            number_of_cliffords=number_of_cliffords,
            apply_inverse=apply_inverse,
            interleaved_clifford_idx=interleaved_clifford_idx,
            seed=seed,
        )

        circuit = self._gate_decomposition(
            clifford_indices=clifford_indices, 
            interleaved_clifford_gate=interleaved_clifford_gate,
        )
        return circuit


import numpy as np
from copy import deepcopy
from typing import Optional, Union
from qutip import destroy, tensor, basis
from qutip_qip.device import Model

__all__ = ["SarimnerModel"]


class SarimnerModel(Model):
    """
    Initializes a new quantum system simulation configuration for the Sarimner model.

    This class sets up the essential parameters and structure needed for simulating a quantum system
    with specified transmon qubit characteristics and couplings. It initializes the internal state
    required for managing the system's dynamics, including drift and controls.

    Parameters
    ----------
    transmon_dict : dict
        A dictionary containing the parameters for each transmon qubit. The keys are qubit
        identifiers, and the values are dictionaries with qubit properties (e.g., frequency,
        anharmonicity).
    coupling_dict : Optional[dict], default=None
        A dictionary specifying the couplings between qubits. If None, no couplings are considered.
    dim : int, default=3
        The dimension of the Hilbert space for each qubit (default is 3 for qutrit).

    Attributes
    ----------
    num_qubits : int
        The number of qubits in the system.
    dims : list of int
        A list specifying the dimension of each qubit's Hilbert space.
    params : dict
        A dictionary containing all system parameters, including transmon properties and couplings.
    _drift : list
        Internal representation of the system's drift Hamiltonian.
    _controls : dict
        Internal setup for system control Hamiltonians.
    _noise : list
        An empty list initialized for potential future addition of noise models.

    Methods
    -------
    _parse_dict(input_dict: dict) -> dict
        Internal method to parse and validate input dictionaries.
    _set_up_drift() -> list
        Internal method to set up the drift Hamiltonian.
    _set_up_controls() -> dict
        Internal method to set up the control Hamiltonians.

    Notes
    -----
    - The `transmon_dict` should contain parameters for each qubit, such as frequency and anharmonicity.
    - The `coupling_dict`, if provided, should specify the coupling strengths between qubit pairs.
    - The model uses a fixed dimension (`dim`) for all qubits, defaulting to qutrit (3-level) systems.
    """
    def __init__(self, transmon_dict: dict, coupling_dict: Optional[dict] = None, dim: int = 3):
        self.num_qubits = int(len(transmon_dict))
        # dimension of each subsystem
        self.dims = [dim] * self.num_qubits
        self.params = {
            "transmons": self._parse_dict(transmon_dict),
            "couplings": self._parse_dict(coupling_dict) if coupling_dict is not None else None,
        }

        # setup drift, controls an noise
        self._drift = self._set_up_drift()
        self._controls = self._set_up_controls()
        self._noise = []

    @staticmethod
    def _parse_dict(d: dict) -> dict:
        """Multiply the values of the dict with 2*pi.

        Args:
            d (dict): dictionary with frequencies.

        Returns:
            dict: dictionary where the frequencies that have been converted to radial frequencies.
        """
        def multiply_values(d: dict):
            for key, value in d.items():
                if isinstance(value, dict):
                    multiply_values(value)
                elif isinstance(value, (int, float)):
                    # Multiply value by 2*pi to get radial frequency.
                    d[key] = 2 * np.pi * value
        # Create copies of the dictionaries to avoid modifying the original one
        new_dict = deepcopy(d)
        # Apply multiplication
        multiply_values(new_dict)
        return new_dict

    def _set_up_drift(self):
        drift = []
        for key, value in self.params["transmons"].items():
            destroy_op = destroy(self.dims[key])
            alpha = value["anharmonicity"]
            # We are simulating qubits in the rotating frame
            drift.append((alpha / 2 * destroy_op.dag()**2 * destroy_op**2, [key]))
        return drift

    def _set_up_controls(self):
        """
        Generate the Hamiltonians and save them in the attribute `controls`.
        """
        dims = self.dims
        controls = {}

        for key in self.params["transmons"].keys():
            destroy_op = destroy(dims[key])
            controls["x" + str(key)] = (destroy_op.dag() + destroy_op, [key])
            controls["y" + str(key)] = (1j*(destroy_op.dag() - destroy_op), [key])

        if self.params["couplings"] is not None:
            for (key1, key2), value in self.params["couplings"].items():
                # Create basis states
                ket01 = tensor(basis(self.dims[key1],0), basis(self.dims[key2],1))
                ket10 = tensor(basis(self.dims[key1],1), basis(self.dims[key2],0))
                ket11 = tensor(basis(self.dims[key1],1), basis(self.dims[key2],1))
                ket20 = tensor(basis(self.dims[key1],2), basis(self.dims[key2],0))

                g = value # coupling strength
                iswap_op = g * (ket01*ket10.dag() + ket10*ket01.dag())
                cz_op_real = np.sqrt(2) * g * (ket11*ket20.dag() + ket20*ket11.dag())
                cz_op_imag = 1j * np.sqrt(2) * g * (ket11*ket20.dag() - ket20*ket11.dag())

                controls["iswap" + str(key1) + str(key2)] = (iswap_op, [key1, key2])
                controls["cz_real" + str(key1) + str(key2)] = (cz_op_real, [key1, key2])
                controls["cz_imag" + str(key1) + str(key2)] = (cz_op_imag, [key1, key2])

        return controls

    def get_control_latex(self):
        """
        Get the labels for each Hamiltonian.
        It is used in the method :meth:`.Processor.plot_pulses`.
        It is a 2-d nested list, in the plot,
        a different color will be used for each sublist.
        """
        num_qubits = self.num_qubits
        labels = [
            {f"x{n}": "$sx_{" + f"{n}" + "}$" for n in range(num_qubits)},
            {f"y{n}": "$sy_{" + f"{n}" + "}$" for n in range(num_qubits)},
        ]
        label_zz = {}

        for m in range(num_qubits - 1):
            for n in range(m + 1, num_qubits):
                label_zz[f"iswap{m}{n}"] = r"$iswap_{"+f"{m}{n}"+"}$"
                label_zz[f"cz_real{m}{n}"] = r"$cz_{" + f"{m}{n}" + "}$"
                label_zz[f"cz_imag{m}{n}"] = r"$\Im cz_{" + f"{m}{n}" + "}$"

        labels.append(label_zz)
        return labels



import unittest
import numpy as np
from qutip import basis, destroy, Options, Qobj, fidelity, tensor
from qutip_qip.circuit import QubitCircuit
from chalmers_qubit.devices.sarimner import SarimnerModel, SarimnerProcessor, DecoherenceNoise, ZZCrossTalk
from chalmers_qubit.utils.operations import project_on_qubit


def _bloch_redfield(psi: Qobj, t1, t2, t: float) -> Qobj:
    psi_data = psi.full()
    alpha = psi_data[0, 0]
    beta = psi_data[1, 0]
    Gamma_1 = 1 / t1
    Gamma_2 = 1 / t2
    rho_00 = 1 + (abs(alpha) ** 2 - 1) * np.exp(-Gamma_1 * t)
    rho_01 = alpha * np.conj(beta) * np.exp(-Gamma_2 * t)
    rho_10 = np.conj(alpha) * beta * np.exp(-Gamma_2 * t)
    rho_11 = abs(beta) ** 2 * np.exp(-Gamma_1 * t)
    rho = np.array([[rho_00, rho_01], [rho_10, rho_11]])
    return Qobj(rho)

class TestSingleQubitNoise(unittest.TestCase):
    def setUp(self) -> None:
        transmon_dict = {
            0: {"frequency": 5.0, "anharmonicity": -0.30},
        }
        self.decoherence_dict = {
            0: {"t1": 50e3, "t2": 40e3},
        }
        # Create the processor with the given hardware parameters
        # Load the physical parameters onto the model
        self.model = SarimnerModel(transmon_dict=transmon_dict)
        return super().setUp()

    def tearDown(self) -> None:
        return super().tearDown()

    def _spin_echo(self, tau:float):
        # Circuit for measuring Hahn-echo
        circuit = QubitCircuit(1)
        circuit.add_gate("RX", targets=0, arg_value=np.pi / 2)
        circuit.add_gate("IDLE", targets=0, arg_value=tau / 2)
        circuit.add_gate("RX", targets=0, arg_value=np.pi)
        circuit.add_gate("IDLE", targets=0, arg_value=tau / 2)
        circuit.add_gate("RX", targets=0, arg_value=np.pi / 2)
        return circuit

    def test_bloch_redfield(self):
        init_state = (basis(3,0) + basis(3,1)).unit()
        t_total = 80*1e3 # ns
        tlist = np.linspace(0, t_total, 10)
        # Add noise
        noise = [DecoherenceNoise(self.decoherence_dict)]
        # Create the processor with the given hardware parameters
        sarimner = SarimnerProcessor(model=self.model, noise=noise)
        # Master equation simulation
        result = sarimner.run_state(init_state, tlist=tlist)
        final_state = project_on_qubit(result.states[-1])
        t1 = self.decoherence_dict[0]["t1"]
        t2 = self.decoherence_dict[0]["t2"]
        predicted_state = _bloch_redfield(init_state, t1, t2, t_total)
        f = fidelity(final_state, predicted_state)
        self.assertAlmostEqual(1, f, places=6)

    def test_relaxation(self):
        # Add noise
        noise = [DecoherenceNoise(self.decoherence_dict)]
        # Create the processor with the given hardware parameters
        sarimner = SarimnerProcessor(model=self.model, noise=noise)
        # master equation simulation
        init_state = basis(3, 1)
        a = destroy(3)
        e_ops = [a.dag() * a]
        t_total = 200*1e3 # ns
        tlist = np.linspace(0, t_total, 30)
        result = sarimner.run_state(init_state, tlist=tlist, e_ops=e_ops)
        # Get the population of the |0> state
        simulated_relaxation = result.expect[0]
        t1 = self.decoherence_dict[0]["t1"]
        theoretical_relaxation = np.exp(-tlist / t1)
        # Verify that simulated relaxation follows the theoretical prediction within a tolerance
        np.testing.assert_allclose(
            simulated_relaxation,
            theoretical_relaxation,
            rtol=0.05,
            atol=0.0,
            err_msg="Simulated relaxation does not match theoretical prediction.",
        )

    def test_spin_echo(self):
        # Do spin echo / Hahn echo experiment
        noise = [DecoherenceNoise(self.decoherence_dict)]  # Create noise
        # Initial state for the simulation
        init_state = basis(3, 0)
        # Idling time list in (ns)
        idle_tlist = np.linspace(0,100*1e3,5)
        # Expectation value of number operator
        e_ops = [destroy(3).dag()*destroy(3)]
        # Create the processor with the given hardware parameters
        sarimner = SarimnerProcessor(model=self.model, noise=noise)
        # Population list
        simulated_decay = []
        # Total time list
        total_time = []
        # Loop over times
        for tau in idle_tlist:
            times, coeffs = sarimner.load_circuit(self._spin_echo(tau))
            # Get the total time of the simulation
            t_total = sarimner.get_full_tlist()[-1]
            # Run master equation simulation and save only start and end times
            result = sarimner.run_state(init_state, tlist=[0,t_total] , e_ops=e_ops, options=Options(nsteps=5e6))
            # Save results
            simulated_decay.append(result.expect[0][-1])
            total_time.append(t_total)
        # Theoretical decay for spin echo
        t2 = self.decoherence_dict[0]["t2"]
        theoretical_decay = (np.exp(-np.array(total_time) / t2) + 1) / 2
        np.testing.assert_allclose(
            1-np.array(simulated_decay),
            theoretical_decay,
            rtol=0.05,
            atol=0.0,
            err_msg="Simulated decay does not match theoretical prediction.",
        )


class TestTwoQubitNoise(unittest.TestCase):
    def setUp(self) -> None:
        transmon_dict = {
            0: {"frequency": 5.0, "anharmonicity": -0.30},
            1: {"frequency": 5.4, "anharmonicity": -0.30},
        }
        coupling_dict = {(0,1): 1e-3}
        self.model = SarimnerModel(
            transmon_dict=transmon_dict,
            coupling_dict=coupling_dict
        )
        # Cross talk matrix (GHz)
        self.cross_talk_dict = {(0,1): 2e-4}
        return super().setUp()

    def tearDown(self) -> None:
        return super().tearDown()

    def test_zz_cross_talk(self):
        # Add noise
        noise = [ZZCrossTalk(self.cross_talk_dict)]
        # Create the processor with the given hardware parameters
        sarimner = SarimnerProcessor(model=self.model, noise=noise)
        # Prepare qubit in 11 state
        rho = tensor(basis(3, 1), basis(3, 1))
        # Transmon Hamiltonian with a slight detuning
        t_total = 1000  # in (ns)
        tlist = np.linspace(0, t_total, 50)
        # Master equation simulation
        result = sarimner.run_state(rho, tlist=tlist)
        final_state = project_on_qubit(result.states[-1])
        simulated_phase = np.angle(final_state.full()[3,0])
        zz = 2*np.pi*self.cross_talk_dict[(0,1)]
        predicted_phase = np.angle(np.exp(-1j * zz * t_total))
        # Check that predicted phase is equal to simulated phase
        self.assertAlmostEqual(predicted_phase, simulated_phase, places=5)

if __name__ == "__main__":
    unittest.main()



# Modifications copyright (c) 2025 Pontus Vikstål
#
# MIT License
# 
# Copyright (c) 2016 DiCarlo lab-QuTech-Delft University of Technology
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from typing import Literal

I = np.eye(4)

# Pauli group
X = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, -1, 0], [0, 0, 0, -1]], dtype=int)
Y = np.array([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]], dtype=int)
Z = np.array([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]], dtype=int)

# Exchange group
S = np.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 1, 0, 0], [0, 0, 1, 0]], dtype=int)
S2 = np.dot(S, S)

# Hadamard group
H = np.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 0, -1, 0], [0, 1, 0, 0]], dtype=int)

# CZ
CZ = np.array(
    [
        [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
        [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
    ],
    dtype=int,
)

def X_theta(theta: float, unit: Literal["deg", "rad"] = "deg") -> np.ndarray:
    """
    Return the Pauli Transfer Matrix (PTM) of a rotation of angle theta
    around the X-axis.

    Args:
        theta (float): Rotation angle.
        unit (str): Unit of the angle, either "deg" for degrees or "rad" for radians.

    Returns:
        np.ndarray: The 4x4 PTM matrix corresponding to the X rotation.
    """
    if unit == "deg":
        theta = np.deg2rad(theta)
    elif unit != "rad":
        raise ValueError(f"Unsupported unit '{unit}'. Use 'deg' or 'rad'.")
    
    cos = np.cos(theta)
    sin = np.sin(theta)

    X = np.array(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, cos, -sin],
            [0, 0, sin, cos],
        ],
        dtype=np.float64,
    )
    return X


def Y_theta(theta: float, unit: Literal["deg", "rad"] = "deg") -> np.ndarray:
    """
    Return the Pauli Transfer Matrix (PTM) of a rotation of angle theta
    around the Y-axis.

    Args:
        theta (float): Rotation angle.
        unit (str): Unit of the angle, either "deg" for degrees or "rad" for radians.

    Returns:
        np.ndarray: The 4x4 PTM matrix corresponding to the Y rotation.
    """
    if unit == "deg":
        theta = np.deg2rad(theta)
    elif unit != "rad":
        raise ValueError(f"Unsupported unit '{unit}'. Use 'deg' or 'rad'.")

    cos = np.cos(theta)
    sin = np.sin(theta)

    Y = np.array(
        [
            [1, 0,    0,   0],
            [0, cos,  0,  sin],
            [0, 0,    1,   0],
            [0, -sin, 0,  cos],
        ],
        dtype=np.float64,
    )
    return Y


import unittest
import numpy as np
from qutip import qeye, average_gate_fidelity
from qutip_qip.operations import CZ, H

from chalmers_qubit.utils import RandomizedBenchmarking
from chalmers_qubit.utils.randomized_benchmarking.clifford_group import SingleQubitClifford, TwoQubitClifford

class TestRandomizedBenchmarking(unittest.TestCase):

    def setUp(self) -> None:
        # Test cases for different Clifford groups
        self.test_cases = [
            (1, H(0)),  # (clifford_group, interleaved_clifford_idx)
            (2, CZ(controls=0, targets=1)) # Example interleaved indices
        ]

    def test_init(self):
        """Test RandomizedBenchmarking initialization."""
        # Test valid initialization
        rb1 = RandomizedBenchmarking(clifford_group=1)
        self.assertEqual(rb1.clifford_group, 1)
        self.assertEqual(rb1.CliffordClass, SingleQubitClifford)
        
        rb2 = RandomizedBenchmarking(clifford_group=2)
        self.assertEqual(rb2.clifford_group, 2)
        self.assertEqual(rb2.CliffordClass, TwoQubitClifford)
        
        # Test invalid initialization
        with self.assertRaises(NotImplementedError):
            RandomizedBenchmarking(clifford_group=3)

    def test_randomized_benchmarking_sequence(self):
        """Test that the randomized benchmarking sequence is generated correctly."""
        for clifford_group, _ in self.test_cases:
            with self.subTest(clifford_group=clifford_group):
                rb = RandomizedBenchmarking(clifford_group=clifford_group)
                number_of_cliffords = 100
    
                # Test without inverse
                sequence = rb._randomized_benchmarking_sequence(
                    number_of_cliffords=number_of_cliffords,
                    apply_inverse=False,
                    interleaved_clifford_idx=None,
                    seed=123
                )
                self.assertEqual(len(sequence), number_of_cliffords)
                
                # Test with inverse
                sequence_with_inverse = rb._randomized_benchmarking_sequence(
                    number_of_cliffords=number_of_cliffords,
                    apply_inverse=True,
                    interleaved_clifford_idx=None,
                    seed=123
                )
                self.assertEqual(len(sequence_with_inverse), number_of_cliffords + 1)


    def test_recovery_clifford_produces_identity(self):
        """Test that recovery gate in clifford sequence yields identity."""
        for clifford_group, _ in self.test_cases:
            with self.subTest(clifford_group=clifford_group):
                rb = RandomizedBenchmarking(clifford_group=clifford_group)
                sequence = rb._randomized_benchmarking_sequence(
                    number_of_cliffords=10,  # Using smaller number for faster testing
                    apply_inverse=True,
                    seed=123
                )
                
                # Get the recovery Clifford (last element)
                recovery_clifford = rb.CliffordClass(sequence[-1])
                
                # Compute net Clifford from the sequence (excluding recovery)
                net_clifford = rb._calculate_net_clifford(sequence[:-1])
                
                # Compute the product of the net clifford and the recovery clifford
                total_clifford = net_clifford * recovery_clifford
                identity_clifford = rb.CliffordClass(0)  # Identity is at index 0
                
                self.assertTrue(
                    total_clifford == identity_clifford, 
                    "The product of a sequence and its recovery should be the identity"
                )

    def test_interleaved_randomized_benchmarking(self):
        """Test that recovery gate in interleaved clifford sequence yields identity."""
        for clifford_group, interleaved_clifford_gate in self.test_cases:
            with self.subTest(clifford_group=clifford_group):
                rb = RandomizedBenchmarking(clifford_group=clifford_group)
                interleaved_clifford_idx = rb._get_interleaved_clifford_idx(interleaved_clifford_gate)
                sequence = rb._randomized_benchmarking_sequence(
                    number_of_cliffords=10,  # Using smaller number for faster testing
                    apply_inverse=True,
                    interleaved_clifford_idx=interleaved_clifford_idx,
                    seed=123
                )
                
                # Get the recovery Clifford (last element)
                recovery_clifford = rb.CliffordClass(sequence[-1])
                
                # Compute net Clifford from the sequence (excluding recovery)
                net_clifford = rb._calculate_net_clifford(sequence[:-1])
                
                # Compute the product of the net clifford and the recovery clifford
                total_clifford = net_clifford * recovery_clifford
                identity_clifford = rb.CliffordClass(0)  # Identity is at index 0
                
                self.assertTrue(
                    total_clifford == identity_clifford, 
                    "The product of an interleaved sequence and its recovery should be the identity"
                )

    def test_randomized_benchmarking_circuit(self):
        """Test that the randomized benchmarking circuit produces high fidelity."""
        for clifford_group, _ in self.test_cases:
            with self.subTest(clifford_group=clifford_group):
                rb = RandomizedBenchmarking(clifford_group=clifford_group)
                
                # Generate a short RB circuit for testing
                circuit = rb.randomized_benchmarking_circuit(
                    number_of_cliffords=5,  # Short sequence for faster testing
                    apply_inverse=True,
                    seed=123
                )
                
                # Check that circuit has the correct number of qubits
                self.assertEqual(circuit.N, clifford_group)
                
                # Compute unitary and check fidelity with identity
                U = circuit.compute_unitary()
                identity = qeye(dimensions=[2]*clifford_group)
                
                f = average_gate_fidelity(U, identity)
                self.assertAlmostEqual(
                    1, f, places=10, 
                    msg="RB circuit should have high fidelity with identity when recovery is applied"
                )

if __name__ == '__main__':
    unittest.main()


import numpy as np
from typing import Optional
from qutip import destroy, num, tensor
from qutip_qip.pulse import Pulse, Drift
from qutip_qip.noise import Noise

__all__ = ["DecoherenceNoise", "ZZCrossTalk"]


class DecoherenceNoise(Noise):
    """
    Represents decoherence noise in a quantum system, characterized by T1 and T2 times for each qubit.

    This class models the decoherence effects on qubits in a quantum system. It allows for
    specifying individual T1 (relaxation) and T2 (dephasing) times for each qubit in the system.

    Parameters
    ----------
    decoherence_dict : dict
        A dictionary specifying the decoherence parameters for each qubit. The keys should be
        qubit identifiers, and the values should be dictionaries containing 't1' and 't2' keys
        with their respective time values in nanoseconds.

    Attributes
    ----------
    decoherence : dict
        A dictionary storing the T1 and T2 values for each qubit in the system.

    Example
    -------
    ```python
    decoherence_dict = {
        0: {'t1': 50e3, 't2': 70e3}, # values for the 0th qubit
        1: {'t1': 45e3, 't2': 60e3} # values for the 1st qubit
    }
    noise = [DecoherenceNoise(decoherence_dict)]
    ```

    Notes
    -----
    - T1 represents the relaxation time, which characterizes the time scale for energy dissipation.
    - T2 represents the dephasing time, which characterizes the time scale for loss of phase coherence.
    - T2 is always less than or equal to 2*T1.
    """

    def __init__(self, decoherence_dict:dict):
        self.decoherence = decoherence_dict

    def get_noisy_pulses(self, dims:list, pulses:Optional[Pulse]=None, systematic_noise:Optional[Pulse]=None):
        """
        Return the input pulses list with noise added and
        the pulse independent noise in a dummy :class:`.Pulse` object.

        Parameters
        ----------
        dims: list
            The dimension of the components system.
        pulses : list of :class:`.Pulse`
            The input pulses. The noise will be added to pulses in this list.
        systematic_noise : :class:`.Pulse`
            The dummy pulse with no ideal control element.

        Returns
        -------
        noisy_pulses: list of :class:`.Pulse`
            Noisy pulses.
        systematic_noise : :class:`.Pulse`
            The dummy pulse representing pulse-independent noise.
        """
        if systematic_noise is None:
            systematic_noise = Pulse(None, None, label="system")

        for key, value in self.decoherence.items():
            t1 = value["t1"]
            t2 = value["t2"]
            if t1 is not None:
                op = 1 / np.sqrt(t1) * destroy(dims[key])
                systematic_noise.add_lindblad_noise(op, key, coeff=True)
            if t2 is not None:
                # Keep the total dephasing ~ exp(-t/t2)
                if t1 is not None:
                    if 2 * t1 < t2:
                        raise ValueError(
                            "t1={}, t2={} does not fulfill " "2*t1>t2".format(t1, t2)
                        )
                    T2_eff = 1.0 / (1.0 / t2 - 1.0 / 2.0 / t1)
                else:
                    T2_eff = t2
                op = 1 / np.sqrt(2 * T2_eff) * 2 * num(dims[key])
                systematic_noise.add_lindblad_noise(op, key, coeff=True)
        return pulses, systematic_noise


class ZZCrossTalk(Noise):
    """
    A noise model representing always-on ZZ cross talk between qubits,
    characterized by a cross-talk coefficient for each pair of qubits.

    Parameters
    ----------
    cross_talk_dict : dict
        A dictionary specifying the cross-talk strength between qubits.
        The keys are tuples (i, j), where `i` and `j` are qubit indices,
        and the corresponding value is the ZZ interaction strength
        between qubit `i` and qubit `j`.

    Attributes
    ----------
    cross_talk_dict : dict
        A dictionary representing the ZZ cross-talk interaction strengths
        between qubit pairs.

    Notes
    -----
    - This model assumes that cross talk is always on, meaning it
      continuously affects the qubits throughout their operation.

    Example
    -------
    ```python	
    cross_talk_dict = {
        (0, 1): 0.01,  # ZZ interaction strength between qubit 0 and qubit 1
        (1, 2): 0.02,  # ZZ interaction strength between qubit 1 and qubit 2
    }
    noise = [ZZCrossTalk(cross_talk)]
    ```
    """

    def __init__(self, cross_talk_dict: dict):
        self.cross_talk_dict = cross_talk_dict

    def get_noisy_pulses(self, dims:list, pulses:Optional[Pulse]=None, systematic_noise:Optional[Pulse]=None):
        """
        Return the input pulses list with noise added and
        the pulse independent noise in a dummy :class:`.Pulse` object.

        Parameters
        ----------
        dims: list
            The dimension of the components system, the default value is
            [3,3...,3] for qutrit system.
        pulses : list of :class:`.Pulse`
            The input pulses. The noise will be added to pulses in this list.
        systematic_noise : :class:`.Pulse`
            The dummy pulse with no ideal control element.

        Returns
        -------
        pulses: list of :class:`.Pulse`
            Noisy pulses.
        systematic_noise : :class:`.Pulse`
            The dummy pulse representing pulse-independent noise.
        """
        if systematic_noise is None:
            systematic_noise = Pulse(None, None, label="system")

        for (key1, key2), value in self.cross_talk_dict.items():
            d1 = dims[key1]
            d2 = dims[key2]

            zz_op = tensor(num(d1), num(d2))
            zz_coeff = 2*np.pi*value

            systematic_noise.add_control_noise(
                zz_coeff * zz_op,
                targets=[key1, key2],
                tlist=None,
                coeff=True,
            )
        return pulses, systematic_noise
